from ursina import color
from ursina.shader import Shader
from ursina.texture_importer import load_texture
from ursina.ursinastuff import Func
from ursina.vec3 import Vec3
import time


def _time():
    return time.time() % 2

painterly_shader = Shader(language=Shader.GLSL, fragment='''
#version 140

vec3 sphere[16] = vec3[](
    vec3( 0.5381, 0.1856,-0.4319), vec3( 0.1379, 0.2486, 0.4430),
    vec3( 0.3371, 0.5679,-0.0057), vec3(-0.6999,-0.0451,-0.0019),
    vec3( 0.0689,-0.1598,-0.8547), vec3( 0.0560, 0.0069,-0.1843),
    vec3(-0.0146, 0.1402, 0.0762), vec3( 0.0100,-0.1924,-0.0344),
    vec3(-0.3577,-0.5301,-0.4358), vec3(-0.3169, 0.1063, 0.0158),
    vec3( 0.0103,-0.5869, 0.0046), vec3(-0.0897,-0.4940, 0.3287),
    vec3( 0.7119,-0.0154,-0.0918), vec3(-0.0533, 0.0596,-0.5411),
    vec3( 0.0352,-0.0631, 0.5460), vec3(-0.4776, 0.2847,-0.0271)
);

uniform sampler2D tex;
uniform sampler2D dtex;
uniform sampler2D random_texture;
uniform mat4 p3d_ViewProjectionMatrix;

in vec2 uv;
out vec4 o_color;

uniform float numsamples;
uniform float radius;
uniform float amount;
uniform float strength;
uniform float falloff;
uniform float time; // add this to drive scrolling noise

vec3 get_normal(vec2 texcoords) {
    const vec2 offset1 = vec2(0.0, 0.001);
    const vec2 offset2 = vec2(0.001, 0.0);

    float depth = texture(dtex, texcoords).r;
    float depth1 = texture(dtex, texcoords + offset1).r;
    float depth2 = texture(dtex, texcoords + offset2).r;

    vec3 p1 = vec3(offset1, depth1 - depth);
    vec3 p2 = vec3(offset2, depth2 - depth);

    vec3 normal = cross(p1, p2);
    normal.z = -normal.z;

    return normalize(normal);
}

vec3 reconstructPosition(in vec2 uv, in float z) {
    float x = uv.x * 2.0 - 1.0;
    float y = (1.0 - uv.y) * 2.0 - 1.0;
    vec4 position_s = vec4(x, y, z, 1.0);
    mat4x4 view_projection_matrix_inverse = inverse(p3d_ViewProjectionMatrix);
    vec4 position_v = view_projection_matrix_inverse * position_s;
    return position_v.xyz / position_v.w;
}

void main() {
    float depth = texture(dtex, uv).r;
    vec3 position = reconstructPosition(uv, depth);
    vec3 normal = get_normal(uv);

    vec3 baseColor = texture(tex, uv).rgb;
    vec3 scatteredColor = vec3(0.0);
    float totalWeight = 0.0;

    for(int i = 0; i < int(numsamples); ++i) {
        vec2 noiseScroll = vec2(time * 0.05, time * 0.07); // scroll speed, adjust as needed
        vec3 random_vector = texture(random_texture, uv * (float(i) + 1.0) * 50.0 + noiseScroll).xyz * 2.0 - 1.0;
        vec3 ray = (radius / depth) * reflect(sphere[i], random_vector);

        vec2 offsetUV = uv + ray.xy;
        float neighborDepth = texture(dtex, offsetUV).r;
        float depth_difference = abs(depth - neighborDepth);

        float weight = smoothstep(0.0, falloff, depth_difference) * strength;

        vec3 sampleColor = texture(tex, offsetUV).rgb;
        scatteredColor += sampleColor * weight;
        totalWeight += weight;
    }

    if(totalWeight > 0.0) {
        scatteredColor /= totalWeight;
    } else {
        scatteredColor = baseColor;
    }

    vec3 color = baseColor;
    float luminance = dot(color, vec3(0.299, 0.587, 0.114));
    float neighborLuminance = dot(texture(tex, uv + vec2(0.001, 0)).rgb, vec3(0.299, 0.587, 0.114));
    float luminanceContrast = abs(luminance - neighborLuminance);

    float neighborDepth = texture(dtex, uv + vec2(0.001, 0)).r;
    float depthContrast = abs(depth - neighborDepth);

    float contrast = max(luminanceContrast * 2.0, depthContrast * 10.0);
    float contrastMask = smoothstep(0.02, 0.05, contrast);

    o_color.rgb = mix(baseColor, scatteredColor, clamp(amount * contrastMask, 0.0, 1.0));
    o_color.a = 1.0;
}
''',

default_input = {
    'numsamples' : 1,
    'radius' : 0.0025,
    'amount' : .8,
    'strength' : .1,
    'falloff' : 0.5,
    'random_texture' : Func(load_texture, 'perlin_noise'),
    'clip_plane_near' : .01,
    'time':0.0,
},
# continuous_input = {
#     'time': _time,
# }
)

if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere', color=color.orange)
    e = Entity(model='cube', y=-1)
    e = Entity(model='plane', scale=100, y=-1)
    Sky()
    Button(y=-.4, scale=.1)
    camera.shader = painterly_shader
    noise_tex = load_texture('noise')
    noise_tex.repeat = True

    EditorCamera()
    def update():
        print(time.time() % 1)

    def input(key):
        if key == 'space':
            if camera.shader:
                camera.shader = None
            else:
                camera.shader = painterly_shader


    random.seed(2)
    for i in range(20):
        e = Entity(model='cube', position=Vec3(random.random(),random.random(),random.random())*3, rotation=Vec3(random.random(),random.random(),random.random())*360)
        # e.shader = matcap_shader


    app.run()



from ursina import *

tesselation_shader = Shader(language=Shader.GLSL,
vertex = '''
#version 400

in vec4 p3d_Vertex;
out vec3 controlpoint_wor;

void main() {
    controlpoint_wor = p3d_Vertex.xyz;
}
''',
tessControl='''

// Compute the tesselation factors here
layout(vertices = 3) out;
in vec3 controlpoint_wor[];
out vec3 tcPosition[];
uniform float TessLevelInner = 4.0;
uniform float TessLevelOuter = 4.0;

// #define ID gl_InvocationID

void main()
{
    tcPosition[ID] = controlpoint_wor[ID];
    if (ID == 0) {
        gl_TessLevelInner[0] = TessLevelInner;
        gl_TessLevelOuter[0] = TessLevelOuter;
        gl_TessLevelOuter[1] = TessLevelOuter;
        gl_TessLevelOuter[2] = TessLevelOuter;
    }
}
''',

tessEval='''
#version 400

// This shader creates the triangles
layout(triangles, equal_spacing, ccw) in;
in vec3 tcPosition[];
out vec3 tePosition;
out vec3 tePatchDistance;
uniform mat4 p3d_ModelViewProjectionMatrix;

void main()
{
    vec3 p0 = gl_TessCoord.x * tcPosition[0];
    vec3 p1 = gl_TessCoord.y * tcPosition[1];
    vec3 p2 = gl_TessCoord.z * tcPosition[2];
    tePatchDistance = gl_TessCoord;
    tePosition = (p0 + p1 + p2);
    gl_Position = p3d_ModelViewProjectionMatrix * vec4(tePosition, 1);
}
// #endif
''',
fragment='''
// #ifdef _FRAGMENT_
// Final rendering (I know gl_FragColor is deprecated)
void main() {
    gl_FragColor = vec4(1,0,0,1);
}
// #endif

''',
default_input={
    'TessLevelInner' : 8.0,
    'TessLevelOuter' : 8.0,

}
)


if __name__ == '__main__':
    app = Ursina()
    window.color=color.black

    e = Entity(
        model=Circle(3),
        shader=tesselation_shader
    )

    # geomNode = e.model.geomNode
    # for i in range(geomNode.getNumGeoms()):
    #     geomNode.modifyGeom(i).makePatchesInPlace()

    # def input(key):
    #     if key == 'space':
    #         if not camera.shader:
    #             camera.shader = fxaa_shader
    #         else:
    #             camera.shader = None

    EditorCamera()
    app.run()



### tree view ###
from ursina import *
from pathlib import Path


class TreeView(Entity):
    def __init__(self, **kwargs):
        super().__init__(parent=camera.ui, position=window.top_left)
        Text.size = 1/window.size[1]*14
        Text.default_font = 'arial.ttf'
        Text.default_resolution = 14*2
        # self.font = 'arial.ttf'

        self.line_height = 1.5
        self.button_size = (.3, Text.size * self.line_height)
        self.y -= self.button_size[1]

        self.bg = Panel(parent=self, model='quad', origin=(-.5,.5), scale=(self.button_size[0],1), color=color._15, z=1)
        self.path = Path('.')
        self.opened_dirs = list()
        self.opened_dirs.append(self.path / 'Models')
        self.opened_dirs.append(self.path / 'Models' / 'compressed')
        # do this to draw text once for the whole tree view instead of for every button
        self.text = ''
        self.text_entity = Text(parent=self, z=-1, line_height=self.line_height)
        self.text_entity.world_y -= (self.button_size[1] * 20) + self.button_size[1] * 5

        self.top_bar = Button(
            text = ' <gray>* ' + str(self.path.resolve()),
            model = 'quad',
            color = color._30,
            scale = (camera.aspect_ratio, self.button_size[1]),
            text_origin = (-.5,0),
            origin = (-.5, .5),
            pressed_scale = 1,
            position = window.top_left
            )

        self.draw_tree()


    def draw_folder(self, folder):
        # global text
        arrow = '> '
        if folder in self.opened_dirs:
            arrow = 'v '

        # print('    '*str(folder).count('\\') + arrow + folder.name)
        self.text += ' <white>'+'  '*str(folder).count('\\') + arrow + folder.name + '\n'
        b = Button(
            parent = self,
            model = 'quad',
            scale = self.button_size,
            pressed_scale = 1
            )

        def folder_button_click():
            if not folder in self.opened_dirs:
                self.opened_dirs.append(folder)
            else:
                self.opened_dirs.remove(folder)
            self.draw_tree()

        b.on_click = folder_button_click


        if folder in self.opened_dirs:
            for subfolder in [e for e in folder.glob('*') if e.is_dir()]:
                self.draw_folder(subfolder)

            for file in [e for e in folder.glob('*') if e.is_file()]:
                self.text += ' <gray>' + '  '*str(file).count('\\') + file.name + '\n<default>'
                b = Button(
                    parent = self,
                    model = 'quad',
                    scale = self.button_size,
                    pressed_scale = 1
                    )


    def draw_tree(self):
        t = time.time()
        self.text = ''
        [destroy(e) for e in self.children if not e in (self.bg, self.text_entity)]

        self.project_button = Button(
            parent = self,
            text = ' ' + self.path.resolve().name,
            model = 'quad',
            color = color.yellow.tint(-.5),
            scale = self.button_size,
            text_origin = (-.5, 0),
            pressed_scale = 1,
            )

        for folder in [e for e in self.path.glob('*') if e.is_dir()]:
            self.draw_folder(folder)

        for i, e in enumerate([e for e in self.children if not e in (self.bg, self.text_entity)]):
            e.origin = (-.5, .5)
            e.y = -i*self.button_size[1]

        self.text_entity.text = self.text
        print('drew file tree:', time.time() - t)




if __name__ == '__main__':
    app = Ursina()

    TreeView()

    Button.color = color._20
    window.color = color._25
    window.size /= .75
    window.x = 200
    app.run()



from panda3d.core import Shader


camera_contrast_shader = Shader.make('''

void vshader(float4 vtx_position : POSITION,
            float2 vtx_texcoord0 : TEXCOORD0,
            out float4 l_position : POSITION,
            out float2 l_texcoord0 : TEXCOORD0,
            uniform float4 texpad_tex,
            uniform float4x4 mat_modelproj)
{
    l_position=mul(mat_modelproj, vtx_position);
    l_texcoord0 = vtx_position.xz * texpad_tex.xy + texpad_tex.xy;
}

half3 AdjustContrast(half3 color, half contrast) {
    return saturate(lerp(half3(0.5, 0.5, 0.5), color, contrast));
}

void fshader(float2 l_texcoord0 : TEXCOORD0,
             out float4 o_color : COLOR,
             uniform sampler2D k_tex : TEXUNIT0,
             uniform float2 iResolution,
             uniform float k_iTime)
{
    // o_color = float4(k_iTime,0,0,1);
    float2 uv = l_texcoord0/iResolution.xy;
    // float2 uv = l_texcoord0;
    float3 col = 0.5 + 0.5*cos(k_iTime+uv.xyx+float3(0,.4,.8));
    o_color = float4(col, 1.0);
}


''', Shader.SL_Cg)



if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    camera.shader = camera_contrast_shader
    camera.set_shader_input('iTime', 0)
    camera.set_shader_input('iResolution', (1,1))

    t = 0

    def update():
        global t
        t += time.dt
        camera.set_shader_input('iTime', t)

    EditorCamera()

    app.run()



from ursina import *

app = Ursina()


player = Entity(model='cube', color=color.azure, origin_y=-.5)
player.speed = 3
ground = Entity(model='plane', scale=8, color=color.lime, texture='white_cube', texture_scale=(8,8))


camera.position = (4,8,-16)
camera.look_at((0,0,0))


def update():
    player.z += held_keys['w'] * time.dt
    player.z -= held_keys['s'] * time.dt
    player.x += held_keys['d'] * time.dt
    player.x -= held_keys['a'] * time.dt


app.run()



from modulefinder import ModuleFinder
import os
import sys
import shutil
from shutil import copy
from pathlib import Path
import time
from textwrap import dedent
import platform

def build():
    project_folder = Path.cwd()
    project_name = project_folder.stem
    build_folder = Path(project_folder / f'build_{platform.system()}')
    build_folder.mkdir(exist_ok=True)

    ignore_folders = []
    ignore_filetypes = []

    compressed_textures = []
    compressed_textures_folder = Path(project_folder/'textures_compressed')
    if compressed_textures_folder.exists():
        compressed_textures = compressed_textures_folder.iterdir()

    def copytree(src, dst, symlinks=False, ignore_patterns=[], ignore_filetypes=[]):
        src = str(src)
        dst = str(dst)

        for item in os.listdir(src):
            if item in ignore_patterns:
                continue

            s = os.path.join(src, item)
            d = os.path.join(dst, item)
            if os.path.isdir(s):
                try:
                    ignore_pattern = shutil.ignore_patterns(*(ignore_patterns + [f'*{e}' for e in ignore_filetypes]))
                    shutil.copytree(s, d, symlinks, ignore_pattern)
                except Exception as e:
                    print(e)
            else:
                if Path(s).suffix in ignore_filetypes:
                    print('ignore filetype:', Path(s).suffix)
                    continue
                if Path(s).stem in compressed_textures:
                    continue
                shutil.copy2(s, d)



    include_modules = []

    python_dest = Path(build_folder / 'python')
    python_dlls_dest = Path(build_folder / 'python/DLLs')
    python_lib_dest = Path(build_folder / 'python/Lib')
    src_dest = Path(build_folder / 'src')
    build_engine = True
    build_game = True
    compile_to_pyc = True
    entry_point = 'main.py'

    for i, arg in enumerate(sys.argv):
        if arg == '--help':
            print(dedent('''
                package ursina application for windows10.
                provided with project folder path, creates a build folder where
                it copies python and project's dependent packages. requires a main.py file.
                copies game scripts and assets into 'build/src' folder.
                creates a .bat file to start the game.

                --ignore_folders=*      # add assets to ignore, for example: --ignore_folders=temp,unused
                --ignore_filetypes=*    # filetype to ignore, for example: --ignore_filetypes=.blend,.psd
                --name=''               # change project name
                --include_modules=*     # include extra modules like this: --include_modules=module_one,module_two,module_tree
                --overwrite             # don't ask to overwrite existing build, just overwrite
                --skip_engine
                --skip_game
                --compile_to_pyc=True/False

                Make sure to include any extra modules with --include_modules PIL,numpy for example.
                Any errors while the application is running will be logged in log.txt instead of the console.
                '''
                )
            )
            sys.exit()

        elif arg.startswith('--ignore_folders='):
            ignore_folders = arg.split('=')[1].split(',')
            print('ignoring folders:', ignore_folders)

        elif arg.startswith('--ignore_filetypes='):
            ignore_filetypes = arg.split('=')[1].split(',')
            print('ignoring filetypes:', ignore_filetypes)

        elif arg == '--name':
            project_name = sys.argv[i+1]

        elif arg.startswith('--include_modules='):
            include_modules = arg.split('=')[1].split(',')

        elif arg == '--skip_engine':
            build_engine = False
        elif arg == '--skip_game':
            build_game = False

        elif arg == '--compile_to_pyc=True':
            compile_to_pyc = True
        elif arg == '--compile_to_pyc=False':
            compile_to_pyc = False
        elif arg.startswith('--entry_point='):
            entry_point = arg.split('=')[1]


    if (build_engine and python_dest.exists() or (build_game and src_dest.exists())):
        if '--overwrite' not in sys.argv:
            for e in (python_dest, src_dest):
                msg = f'Folder {e} already exists. \nProceed to delete and overwrite?'
                overwrite = input("%s (y/N) " % msg).lower() == 'y'
                # if not overwrite:
                #     print('stopped building')
                #     exit()
                if e == python_dest:
                    build_engine = overwrite
                elif e == src_dest:
                    build_game = overwrite


    print('building project:', project_folder)
    start_time = time.time()

    if build_engine:
        if python_dest.exists():
            shutil.rmtree(str(python_dest))
        python_dest.mkdir()
        python_dlls_dest.mkdir()
        python_lib_dest.mkdir()


        # def copy_python():
        # copy files in python installation folder, but not the folders
        print('copying python')
        python_folder = Path(sys.executable).parent
        [copy(str(f), str(python_dest / f.name)) for f in python_folder.iterdir() if f.is_file()]


        # def copy_python_lib():
        print('copying python Lib files')
        [copy(str(f), str(python_lib_dest / f.name)) for f in Path(python_folder / 'Lib').iterdir() if f.is_file()]


        # def copy_always_included():
        print('copying always included modules and extra modules')
        include_modules = [f'Lib/site-packages/{e}' for e in include_modules]

        always_include = [
            'Lib/collections',
            'Lib/ctypes',
            'Lib/encodings',
            'Lib/importlib',
            'Lib/urllib',
            'Lib/logging',
            'Lib/xml',
            'Lib/re',
            'Lib/json',
            # 'Lib/site-packages/panda3d/',
            'Lib/site-packages/direct/',
            'Lib/site-packages/pyperclip/',
            'Lib/site-packages/PIL/',
            'Lib/site-packages/screeninfo/',
            'DLLs/libffi-7.dll',
            'DLLs/_ctypes.pyd',
            'DLLs',
            'Lib/site-packages/gltf',
            ]

        for path in always_include + include_modules:
            source = python_folder / path
            dest = python_dest / path
            print('copying always include:', source, '-->', dest)

            if source.is_file():
                dest.parent.mkdir(parents=True, exist_ok=True)
                copy(str(source), str(dest))
            elif source.is_dir():
                dest.mkdir(parents=True, exist_ok=True)
                copytree(source, dest)


        print('copying panda3d')
        (python_dest / 'Lib/site-packages/panda3d/').mkdir(parents=True, exist_ok=True)

        for f in (python_folder / 'Lib/site-packages/panda3d/').iterdir():
            if f.name in (
                'avcodec-55.dll',
                'fmodex64.dll',
                'egg.cp39-win_amd64.pyd',
                'libpandaode.dll',
                'ode.cp39-win_amd64.pyd',
                'models',
                'libp3tinydisplay.dll',
                'libp3assimp.dll',
                ):
                continue
            print('copying:', f, '-->', str(python_dest / 'Lib/site-packages/panda3d/' / f.name))
            if f.is_file():
                copy(str(f), str(python_dest / 'Lib/site-packages/panda3d/' / f.name))
            else:
                (python_dest / 'Lib/site-packages/panda3d/' / f.name).mkdir(parents=True, exist_ok=True)
                copytree(f, (python_dest / 'Lib/site-packages/panda3d/' / f.name))

        # def copy_ursina():
        print('copying ursina')
        import importlib
        spec = importlib.util.find_spec('ursina')
        ursina_path = Path(spec.origin).parent
        print('found ursina at:', ursina_path)
        dest = build_folder / 'python/Lib/site-packages/ursina'
        dest.mkdir(parents=True, exist_ok=True)
        copytree(ursina_path, dest, ignore_patterns=['samples', 'unused'], ignore_filetypes=['.blend', '.obj'])


        # print('copying found modules')
        # finder = ModuleFinder()
        # finder.run_script(str(project_folder) + '/main.py')

        # for name, mod in finder.modules.items():
        #     filename = mod.__file__
        #
        #     if filename is None:
        #         continue
        #     if '__' in name:
        #         print('ignore:', filename)
        #         continue
        #
        #     if 'Python' in filename and 'DLLs' in filename:
        #         print('copying:', filename)
        #         copy(filename, str(build_folder / 'python/DLLs'))
        #
        #     elif 'lib\\site-packages\\' in filename:
        #         print('copying module:', filename)
        #         forward_slashed = filename.split('lib\\site-packages\\')[1].replace('\\', '/')
        #         dir = build_folder / 'python/lib/site-packages' / forward_slashed
        #         dir.parent.mkdir(parents=True, exist_ok=True)
        #         copy(filename, dir)



    if build_game:
        if src_dest.exists():
            shutil.rmtree(str(src_dest))
        src_dest.mkdir()

        ignore_folders.extend(['build_Windows', 'build_Linux', 'build.bat', '__pycache__', '.git'])
        ignore_filetypes.extend(['.gitignore', '.psd', '.zip'])

        if compile_to_pyc:
            import py_compile
            for f in project_folder.glob('**/*.py'):
                in_ignore = False
                for e in ignore_folders:
                    if e in [str(e.name) for e  in f.parents]:
                        in_ignore = True
                        print('skip compiling:', f)
                        break
                if in_ignore:
                    continue

                parents = f.relative_to(project_folder).parents
                if 'scenes' in parents:
                    continue
                py_compile.compile(f, src_dest / (str(f)[len(str(project_folder))+1:]+'c'))

            ignore_filetypes.append('.py')


        print('copying assets')
        for f in project_folder.iterdir():
            name = f.name
            dest = Path(src_dest / f.name)
            if name in ignore_folders:
                print('ignore:', f)
                continue
            elif f.is_dir():
                print('copying assetfolder:', f, 'to', dest)
                dest.mkdir(parents=True, exist_ok=True)
                copytree(project_folder / f, dest, ignore_patterns=ignore_folders, ignore_filetypes=ignore_filetypes)
            elif f.is_file():
                if f.suffix in ignore_filetypes:
                    continue
                print('copying asset:', f, 'to', src_dest / f.name)
                copy(str(f), str(dest))


        print('creating .bat file')
        c = ''
        if compile:
            c = 'c'
        with Path(build_folder / f'{project_name}.bat').open('w') as f:
            f.write(dedent(fr'''
                chcp 65001
                set PYTHONIOENCODING=utf-8

                call "python\python.exe" "src\{entry_point}{c}" > "log.txt" 2>&1'''
            ))

    # make exe
    # import subprocess
    # import importlib
    # spec = importlib.util.find_spec('ursina')
    # ursina_path = Path(spec.origin).parent
    # subprocess.call([
    #     f'{ursina_path}\\scripts\\_bat_to_exe.bat',
    #     f'{build_folder}\\{project_name}.bat',
    #     f'\\build\\{project_folder.stem}.exe'
    #     ])


    print('build complete! time elapsed:', time.time() - start_time)


if __name__ == '__main__':
    build()


from ursina import *
shader = Shader(fragment='''
#version 430

uniform sampler2D tex;
in vec2 uv;
uniform float blur_size;
out vec4 color;

void main() {
    float pad = .2;
    vec3 offset = vec3(1.0, 2.0, 3.0);
    color  = texture(tex, uv);
    color += texture(tex, vec2(uv.x - offset.z, uv.y));
    color += texture(tex, vec2(uv.x - offset.y, uv.y));
    color += texture(tex, vec2(uv.x - offset.x, uv.y));
    color += texture(tex, vec2(min(uv.x + offset.x, pad), uv.y));
    color += texture(tex, vec2(min(uv.x + offset.y, pad), uv.y));
    color += texture(tex, vec2(min(uv.x + offset.z, pad), uv.y));
    color /= 7;
    color.w = 1.;
}
''',
# default_input = {
#
# }
)

if __name__ == '__main__':
    app = Ursina()
    window.color = color._16

    e = Entity(model='sphere', color=color.orange)
    e = Entity(model='cube', y=-1)
    e = Entity(model='plane', texture='grass', scale=10, y=-1)
    camera.shader = shader
    EditorCamera()
    app.run()

#     // blend.rgb
# //
# //   This shader converts to black-and-white before calculating
# //   scene brightness.  To do this, it uses a weighted average of
# //   R,G,B.  The blend parameter controls the weighting.
# //
# // desat.x
# //
# //   Desaturation level.  If zero, the bloom's color is equal to
# //   the color of the input pixel.  If one, the bloom's color is
# //   white.
# //
# // trigger.x
# //
# //   Must be equal to mintrigger.
# //
# //   mintrigger is the minimum brightness to trigger a bloom,
# //   and maxtrigger is the brightness at which the bloom
# //   reaches maximum intensity.
# //
# // trigger.y
# //
# //   Must be equal to (1.0/(maxtrigger-mintrigger)) where
# //
# //   mintrigger is the minimum brightness to trigger a bloom,
# //   and maxtrigger is the brightness at which the bloom
# //   reaches maximum intensity.
# //



from ursina import *

class AutoStairs(Entity):
    def __init__(self, height=1, **kwargs):
        super().__init__(**kwargs)
        # self.box = Entity(model='cube', scale=(1,height,.1), origin=(0,.5,0))
        random.seed(0)
        pause_every = 16
        pause_size = 1

        step_height = .1
        step_depth = .15
        steps = math.ceil(height/step_height)
        rail_height = .5


        z = 0
        for i in range(steps):

            step = Entity(
                parent = self,
                # model = 'cube',
                model = Mesh(
                    vertices=(
                        (.5,.5,-.5), (.5,.5,.5), (-.5,.5,.5),
                        (-.5,.5,-.5),(.5,-.5,.5),(-.5,-.5,.5)),
                    triangles=(0,1,2, 0,2,3, 4,5,1, 1,5,2),
                    mode='triangle'
                    ),
                origin = (0,.5,-.5),
                position = (0, -i*step_height, z),
                scale = (1, step_height, step_depth),
                color = color.tint(color.color(0,0,.4), random.uniform(-.05,.05))
                # color = color.color(i*20, 1, .7, .2)
                )

            z += step_depth

            if pause_every and i % pause_every == 0 and i > 0:
                z += pause_size - step_depth
                step.scale_z = pause_size

        # rail poles
        rail_poles = list()
        for i, step in enumerate(self.children):
            for j in range(int(step.scale_z // step_depth)):
                rail_pole = Entity(
                    parent = step,
                    model = 'cube',
                    origin = (0,-.5,0),
                    x = .45,
                    world_z = step.world_z + (j * step_depth * self.scale_z) - (step_depth/2),
                    color = color.dark_gray,
                    world_scale = (.05, rail_height, .05)
                    )
                rail_pole.world_z += step_depth
                rail_poles.append(rail_pole)
                rail_pole_2 = duplicate(rail_pole)
                rail_pole_2.x = -.45


        if len(self.children) == 0:
            return

        verts = list()
        last_step = self.children[-1]
        for step in reversed(self.children):
            verts.append((0, last_step.y-step_height, 0))
            verts.append(step.position + Vec3(0,-step.scale_y,step.scale_z))
            verts.append(step.position + Vec3(0,0,step.scale_z))

            verts.append((0, last_step.y-step_height, 0))
            verts.append(step.position + Vec3(0,0,step.scale_z))
            verts.append(step.position)


        side = Entity(
            parent = self,
            model = Mesh(vertices=verts, mode='triangle', thickness=10),
            x = .5,
            color = color.color(0,0,.35)
            )

        side_2 = duplicate(side, flipped_faces = True)
        side_2.x *= -1

        path = list()
        path.append(rail_poles[0].world_position + (0,rail_height,-.2)) # first
        for rp in rail_poles:
            path.append(rp.world_position + Vec3(0,rail_height,0))
        path.append(rail_poles[len(rail_poles)-1].world_position + (self.forward * .1) + (self.up * (rail_height-.1))) # last

        railing = Entity(model=Pipe(path=path, thicknesses=((.1,.05),)), color=color.orange, world_parent=self)
        railing_2 = Entity(model=Pipe(path=path, thicknesses=((.1,.05),)), color=color.orange, world_parent=self, flipped_faces=True)
        railing_2.scale_x *= -1

        ground_collider = Entity(model='cube', color=color.green,
            parent = self,
            origin = (0,.5,-.5),
            scale_y = .8,
            world_z = step_depth/2
            )
        ground_collider.look_at(last_step)
        ground_collider.scale_z = distance(ground_collider, Vec3(last_step.position) + (Vec3(0,-step_height,step_depth)*2))


        top_collider = Entity(model='cube', color=color.blue,
            parent = self,
            origin = (0,.5,-.5),
            scale_z = step_depth/2,
            )

        rail_collider = duplicate(ground_collider)
        rail_collider.scale_x = .1
        rail_collider.x = .5 - (.1/2)
        rail_collider.scale_y = rail_height * .9
        rail_collider.scale_z = distance(rail_collider, Vec3(last_step.position) - (Vec3(0,-step_height,step_depth)))
        rail_collider.origin_y = -.5

        rail_start = Entity(model='cube', color=color.red,
            x = .5 - (.1/2),
            origin = (0, -.5, -.5),
            scale = (.1, rail_height, step_depth*2)
            )
        rail_end = duplicate(rail_start)
        rail_end.position = (rail_end.x, -steps*step_height, steps*step_depth)
        rail_end.origin_z = .5


if __name__ == '__main__':
    app = Ursina()

    class AutoStairsTester(Entity):
        def __init__(self):
            super().__init__()
            self.height = 1
            self.stair = AutoStairs(height=self.height)

        def input(self, key):
            if key == 'e':
                self.height += .1
                destroy(self.stair)
                self.stair = AutoStairs(height=self.height)
            if key == 'q':
                self.height -= .1
                destroy(self.stair)
                self.stair = AutoStairs(height=self.height)
    # AutoStairs()
    Sky()
    AutoStairsTester()
    EditorCamera()
    camera.orthographic = True
    camera.fov = 2
    # camera.position = Vec3(10, 5, 5) * 1.5
    # camera.position = Vec3(10, 0, 0)
    # camera.look_at((0,0,0))
    origin = Entity(model='sphere', color=color.red, scale=(.01,.01,.01))
    app.run()



import sys
sys.path.append("..")
from ursina import *

class AddScriptButton():

    def __init__(self):
        self.path = None

    def input(self, key):
        if key == 'left mouse down' and self.entity.hovered:
            if self.path:
                scene.editor.selection[0].add_script(os.path.basename(self.path)[:-3])
                # print('scripts:', scene.editor.selection[0].scripts)



from ursina import *

from ursina.scripts.property_generator import generate_properties_for_class
@generate_properties_for_class()
class HealthBar(Button):

    def __init__(self, max_value=100, value=Default, roundness=.25, bar_color=color.red.tint(-.2), highlight_color=color.black66, animation_duration=.1, show_text=True, show_lines=False, text_size=.7, scale=(.5,.025), origin=(-.5,.5), name='health_bar', **kwargs):
        super().__init__(name=name, position=(-.45*window.aspect_ratio,.45), origin=origin, scale=scale, text='hp / max hp', text_size=text_size, radius=roundness, ignore=True)

        self.bar = Entity(parent=self, model=Quad(radius=roundness), origin=origin, z=-.005, color=bar_color, highlight_color=highlight_color, ignore=True)
        self.lines = Entity(parent=self.bar, y=-1, color=color.black33, ignore=True, enabled=show_lines, z=-.05)

        self.max_value = max_value
        self.clamp = True
        self.roundness = roundness
        self.animation_duration = animation_duration
        self.show_lines = show_lines
        self.show_text = show_text
        self.value = self.max_value if value == Default else value

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.text_entity.enabled = show_text


    def value_setter(self, n):
        if self.clamp:
            n = clamp(n, 0, self.max_value)

        self._value = n

        self.bar.animate_scale_x(n/self.max_value, duration=self.animation_duration, curve=curve.in_out_bounce)
        self.text_entity.text = f'{n} / {self.max_value}'

        if self.lines.enabled:
            self.lines.model = Grid(n, 1)
            self.lines.origin = (-.5,-.5)

        if n / self.max_value >= self.scale_y / self.scale_x:
            aspect_ratio = n/self.max_value*self.scale_x / self.scale_y
            self.bar.model = Quad(radius=self.roundness, aspect=aspect_ratio)
        else:
            self.bar.model = 'quad'
        self.bar.origin = self.bar.origin



    def show_text_getter(self):
        return self.text_entity.enabled
    def show_text_setter(self, value):
        self.text_entity.enabled = value

    def show_lines_getter(self):
        return self.lines.enabled
    def show_lines_setter(self, value):
        self.lines.enabled = value

    def bar_color_getter(self):
        return self.bar.color
    def bar_color_setter(self, value):
        self.bar.color = value



    def __setattr__(self, name, value):
        if name == 'scale' and hasattr(self, 'model') and self.model:  # update rounded corners of background when scaling
            self.model.aspect = self.world_scale_x / self.world_scale_y
            self.model.generate()
            if hasattr(self, 'bar'):
                self.bar.model = Quad(aspect=self.world_scale_x/self.world_scale_y, radius=self.roundness)
                self.bar.origin = self.origin

            if hasattr(self, 'text_entity') and self.text_entity:
                self.text_entity.world_scale = 25 * self.text_size

        super().__setattr__(name, value)


if __name__ == '__main__':
    app = Ursina()

    health_bar_1 = HealthBar(bar_color=color.lime.tint(-.25), roundness=.5, max_value=100, value=50, scale=(.5,.1))
    print(health_bar_1.text_entity.enabled, health_bar_1.text_entity.text)
    # health_bar_1.show_text = False
    # health_bar_1.show_lines = True

    def input(key):
        if key == '+' or key == '+ hold':
            health_bar_1.value += 10
        if key == '-' or key == '- hold':
            health_bar_1.value -= 10
            print('ow')
    app.run()



from ursina import *

app = Ursina(size=Vec2(1920,1080)*.5)
window.always_on_top = True
window.position = (0, 1080/2)

entity = Entity(model='cube', texture='white_cube')

def update():
    entity.rotation_y += time.dt * 10
    entity.y = sin(time.time() * 3) * .5
    entity.color = hsv(time.time() * 100, 1, 1)

window.color = hsv(0,0,.1)

Entity(model=Grid(8,8), rotation_x=90, y=-1, scale=8, alpha=.2)
EditorCamera()
app.run()



from panda3d.core import BitMask32, TransformState
from panda3d.bullet import BulletRigidBodyNode, BulletPlaneShape, BulletBoxShape, BulletSphereShape, BulletCylinderShape, BulletCapsuleShape, BulletConeShape, XUp, YUp, ZUp, BulletTriangleMesh, BulletTriangleMeshShape, BulletDebugNode, BulletWorld
from ursina.scripts.property_generator import generate_properties_for_class
from ursina.entity import Entity
from ursina import scene, time
from ursina.vec3 import Vec3

class PhysicsHandler(Entity):
    def __init__(self):
        self._debug_node = BulletDebugNode('Debug')
        self._debug_node.showWireframe(True)
        self._debug_node.showConstraints(True)
        self._debug_node.showBoundingBoxes(False)
        self._debug_node.showNormals(False)
        self.debug_node_path = scene.attachNewNode(self._debug_node)

        self.world = BulletWorld()
        self.world.setGravity(Vec3(0, -9.81, 0))
        self.world.setDebugNode(self.debug_node_path.node())
        super().__init__()

        self.active = False
        self._show_debug = False

    def update(self):
        if self.active:
            self.world.doPhysics(time.dt)
            # print("physics handler now active")

    @property
    def show_debug(self):
        return self._show_debug

    @show_debug.setter
    def show_debug(self, value):
        self._show_debug = value
        if value:
            self.debug_node_path.show()
        else:
            self.debug_node_path.hide()
physics_handler = PhysicsHandler()


class PlaneShape:
    def __init__(self, center=(0,0,0), normal=Vec3(0,1,0), offset=0):
        self.center = center
        self.normal = normal
        self.offset = offset

class BoxShape:
    def __init__(self, center=(0,0,0), size=(1,1,1)):
        self.center = center
        self.size = size

class SphereShape:
    def __init__(self, center=(0,0,0), radius=.5):
        self.center = center
        self.radius = radius

class CylinderShape:
    def __init__(self, center=(0,0,0), radius=.5, height=2, axis='y'):
        self.center = center
        self.radius = radius
        self.height = height
        self.axis = axis

class CapsuleShape:
    def __init__(self, center=(0,0,0), radius=.5, height=2, axis='y'):
        self.center = center
        self.radius = radius
        self.height = height
        self.axis = axis

class ConeShape:
    def __init__(self, center=(0,0,0), radius=.5, height=2, axis='y'):
        self.center = center
        self.radius = radius
        self.height = height
        self.axis = axis

class MeshShape:
    def __init__(self, mesh=None, center=(0,0,0)):
        self.mesh = mesh
        self.center = center


@generate_properties_for_class(getter_suffix='_getter', setter_suffix='_setter')
class _PhysicsBody:
    # Copy animation functions from Entity
    animate = Entity.animate

    animate_position = Entity.animate_position
    animate_x = Entity.animate_x
    animate_y = Entity.animate_y
    animate_z = Entity.animate_z

    animate_rotation = Entity.animate_rotation
    animate_rotation_x = Entity.animate_rotation_x
    animate_rotation_y = Entity.animate_rotation_y
    animate_rotation_z = Entity.animate_rotation_z

    animate_scale = Entity.animate_scale
    animate_scale_x = Entity.animate_scale_x
    animate_scale_y = Entity.animate_scale_y
    animate_scale_z = Entity.animate_scale_z

    _getattr = Entity._getattr
    _setattr = Entity._setattr

    def __init__(self):
        self._visible = False
        self.ignore_paused = False
        self.animations = []

    def __getattr__(self, attribute):
        return getattr(self.node_path.node(), attribute)


    def visible_getter(self):
        return self._visible

    def visible_setter(self, value):
        self._visible = value
        if value:
            self.node_path.show()
        else:
            self.node_path.hide()

    def position_getter(self):
        return Vec3(*self.node_path.getPos())

    def position_setter(self, value):
        self.node_path.setPos(Vec3(value))

    def x_getter(self):
        return self.node_path.getX()

    def x_setter(self, value):
        self.node_path.setX(value)

    def y_getter(self):
        return self.node_path.getY()

    def y_setter(self, value):
        self.node_path.setY(value)

    def z_getter(self):
        return self.node_path.getZ()

    def z_setter(self, value):
        self.node_path.setZ(value)

    def rotation_getter(self):
        rotation = self.node_path.getHpr()
        return Vec3(rotation[1], rotation[0], rotation[2])

    def rotation_setter(self, value):
        self.node_path.setHpr(Vec3(value[1], value[0], value[2]))

    def scale_getter(self):
        scale = self.node_path.getScale()
        return Vec3(scale[0], scale[1], scale[2])

    def scale_setter(self, value):
        value = [e if e!=0 else .001 for e in value]
        self.node_path.setScale(value[0], value[1], value[2])

    def add_force(self, force, point=(0,0,0)):
        self.node_path.node().setActive(True)
        if point != (0,0,0):
            self.node_path.node().applyForce(force, point)
        else:
            self.node_path.node().applyCentralForce(force)

    def apply_impulse(self, force, point=(0,0,0)):
        self.node_path.node().setActive(True)
        if point != (0,0,0):
            self.node_path.node().applyImpulse(force, point)
        else:
            self.node_path.node().applyCentralImpulse(force)

@generate_properties_for_class()
class RigidBody(_PhysicsBody):
    def __init__(self, shape, world=physics_handler.world, entity=None, mass=0, kinematic=False, friction=.5, mask=0x1):
        super().__init__()
        self.world = world
        if self.world is physics_handler.world and not physics_handler.active:
            physics_handler.active = True

        self.rigid_body_node = BulletRigidBodyNode('RigidBody')
        self.rigid_body_node.setMass(mass)
        self.rigid_body_node.setKinematic(kinematic)
        self.friction = friction

        if entity:
            self.node_path = entity.getParent().attachNewNode(self.rigid_body_node)
            self.scale = entity.world_scale
            self.position = entity.world_position
            entity.world_parent = self.node_path
        else:
            self.node_path = render.attachNewNode(self.rigid_body_node)
        self.node_path.node().setIntoCollideMask(BitMask32(mask))

        if isinstance(shape, (list, tuple)):
            for s in shape:
                self.node_path.node().addShape(_convert_shape(s, entity, dynamic=not self.rigid_body_node.isStatic()), TransformState.makePosHpr(s.center, entity.getHpr()))
        else:
            self.node_path.node().addShape(_convert_shape(shape, entity, dynamic=not self.rigid_body_node.isStatic()), TransformState.makePosHpr(shape.center, entity.getHpr()))

        self.world.attachRigidBody(self.node_path.node())
        self.node_path.setPythonTag('Entity', entity)

    def remove(self):
        self.world.removeRigidBody(self.node_path.node())
        self.node_path.removeNode()

    def friction_getter(self):
        return self.rigid_body_node.getFriction()
    def friction_setter(self, value):
        self.rigid_body_node.setFriction(value)

    def lock_rotation_setter(self, value):
        self._lock_rotation = value
        self.rigid_body_node.setAngularFactor(Vec3(*[1-e for e in value]))




def _convert_shape(shape, entity, dynamic=True):
    if isinstance(shape, PlaneShape):
        return BulletPlaneShape(shape.normal, shape.offset)

    elif isinstance(shape, BoxShape):
        return BulletBoxShape(Vec3(shape.size[0]/2, shape.size[1]/2, shape.size[2]/2))

    elif isinstance(shape, SphereShape):
        return BulletSphereShape(shape.radius)

    elif isinstance(shape, MeshShape):
        if entity:
            if shape.mesh is None and entity.model:
                mesh = entity.model
            else:
                mesh = shape.mesh

            geom_target = mesh.findAllMatches('**/+GeomNode').getPath(0).node().getGeom(0)
            output = BulletTriangleMesh()
            output.addGeom(geom_target)

            return BulletTriangleMeshShape(output, dynamic=dynamic)

    elif isinstance(shape, CapsuleShape):
        if shape.axis == 'y':
            axis = YUp
        elif shape.axis == 'z':
            axis = ZUp
        elif shape.axis == 'x':
            axis = XUp
        return BulletCapsuleShape(shape.radius, shape.height-1, axis)

    elif isinstance(shape, CylinderShape):
        if shape.axis == 'y':
            axis = YUp
        elif shape.axis == 'z':
            axis = ZUp
        elif shape.axis == 'x':
            axis = XUp
        return BulletCylinderShape(shape.radius, shape.height, axis)

    elif isinstance(shape, ConeShape):
        if shape.axis == 'y':
            axis = YUp
        elif shape.axis == 'z':
            axis = ZUp
        elif shape.axis == 'x':
            axis = XUp
        return BulletConeShape(shape.radius, shape.height, axis)
    else:
        raise Exception("invalid shape.")



if __name__ == '__main__':
    from ursina import Ursina, Cylinder, Capsule, Cone, Sequence, Func, curve, Wait, EditorCamera

    app = Ursina(borderless=False)

    ground = Entity(model='plane', texture='grass', scale=30)
    RigidBody(shape=PlaneShape(), entity=ground)

    cube = Entity(model='cube', texture='white_cube')
    # cube.rotation = Vec3(0,45,68)
    # cube.scale = Vec3(.5)
    # cube.y = 7
    cube_body = RigidBody(shape=BoxShape(), entity=cube, mass=1)

    # sphere = Entity(model='sphere', texture='brick', y=30)
    # RigidBody(shape=SphereShape(), entity=sphere, mass=5)

    # cylinder = Entity(model=Cylinder(height=2, radius=.5, start=-1), texture='brick', y=17)
    # RigidBody(shape=CylinderShape(height=2, radius=.5), entity=cylinder, mass=3)

    # capsule = Entity(model=Capsule(height=2, radius=.5), texture='brick', y=15)
    # RigidBody(shape=CapsuleShape(height=2, radius=.5), entity=capsule, mass=3)

    # cone = Entity(model=Cone(resolution=8, height=2, radius=.5), texture='brick', y=13)
    # RigidBody(shape=ConeShape(height=2, radius=.5), entity=cone, mass=2)

    # platform = Entity(model='cube', texture='white_cube', y=1, scale=(4,1,4))
    # platform_body = RigidBody(shape=BoxShape(), entity=platform, kinematic=True, friction=1)
    # platform_sequence = Sequence(entity=platform, loop=True)

    # path = [Vec3(-2,1,-2), Vec3(2,1,-2), Vec3(0, 1, 2)]
    # travel_time = 2
    # for target_pos in path:
    #     platform_sequence.append(Func(platform_body.animate_position, value=target_pos, duration=travel_time, curve=curve.linear), regenerate=False)
    #     platform_sequence.append(Wait(travel_time), regenerate=False)
    # platform_sequence.generate()
    # platform_sequence.start()


    # def input(key):
    #     if key == 'space up':
    #         e = Entity(model='cube', texture='white_cube', y=7)
    #         RigidBody(shape=BoxShape(), entity=e, mass=1, friction=1)
    #     if key == 'up arrow':
    #         cube_body.apply_impulse(force=Vec3(0, 10, 0), point=Vec3(0,0,0))
    #         print('impulse applied')

    physics_handler.show_debug = True

    from ursina import *
    class Player(Entity):
        def __init__(self):
            super().__init__(model=Capsule(), color=color.orange, y=2, z=-4)
            self.rb = RigidBody(shape=CapsuleShape(), entity=self, mass=1, kinematic=0, friction=1)
            self.rb.setLinearFactor(Vec3(1,1,1))
            self.rb.lock_rotation = Vec3(1,1,1)


        def update(self):
            self.direction = Vec3(
                self.forward * (held_keys['w'] - held_keys['s'])
                + self.right * (held_keys['d'] - held_keys['a'])
                ).normalized()
            # print(self.direction)
            self.rb.apply_impulse(self.direction*time.dt*200)
            limit = 10

            self.rb.friction = 10 if self.direction.length() <.1 else 1
            print(self.direction.length())
            vel = self.rb.getLinearVelocity()
            self.rb.setLinearVelocity(Vec3(*[clamp(e, -limit, limit) for e in vel]))


    Player()
    ground = Entity(model='plane', scale=10, collider='box', z=10, rotation_x=-30, y=2, color=color.gray, name='ground')
    ground.rb = RigidBody(entity=ground, shape=BoxShape(size=Vec3(1,0,1)))



    EditorCamera()
    app.run()



from panda3d.core import TextNode
from panda3d.core import FontPool
from panda3d.core import Filename

import builtins
import re
import sys

import ursina
from ursina import camera
from ursina import application
from ursina.entity import Entity
from ursina.sequence import Sequence, Func, Wait
from ursina import color
from ursina import destroy
from ursina.shaders.text_shader import text_shader
from ursina.string_utilities import print_warning
from ursina.scripts.property_generator import generate_properties_for_class
# note:
# <scale:n> tag doesn't work well in the middle of text.
# only good for titles for now.

from functools import lru_cache
@lru_cache()
def _search_for_file(name, folders, file_types=None): # prioritizes based on file_type order, then folder order.
    if file_types is None and '.' not in name:
        raise Exception(f'name({name}) has no file type, and no file_types were provided.')

    if '.' in name:
        file_types = ('', )

    for file_type in file_types:
        for folder in folders:
            # print('-----', 'searchpattern:', f'{folder}/**/{name}{file_type}', 'result:', list(folder.glob(f'**/{name}.{file_type}')))
            for file_path in folder.glob(f'**/{name}{file_type}'):
                #print('FOUND FONT:', file_path)
                return file_path

    return None


@generate_properties_for_class()
class Text(Entity):
    size = .025
    default_font = 'OpenSans-Regular.ttf'
    default_monospace_font = 'VeraMono.ttf'
    default_resolution = 1080 * size * 2
    start_tag = '<'
    end_tag = '>'

    def __init__(self, text='', start_tag=start_tag, end_tag=end_tag, ignore=True, **kwargs):
        super().__init__(ignore=ignore)
        self.size = Text.size
        self.parent = camera.ui

        self.setColorScaleOff()
        self.text_nodes = []
        self.images = []
        self.origin = (-.5, .5)

        self.font = Text.default_font
        self.shader = text_shader
        self.shader.compile()
        self.resolution = Text.default_resolution
        self.use_tags = True
        self.line_height = 1
        self.start_tag = start_tag
        self.end_tag = end_tag
        self.text_colors = {'default' : color.text_color}
        self.color = color.text_color

        for color_name in color.color_names:
            self.text_colors[color_name] = color.colors[color_name]

        self.tag = Text.start_tag + 'default' + Text.end_tag
        self.current_color = self.text_colors['default']
        self.scale_override = 1
        self.background_entity = None
        self.appear_sequence = None # gets created when calling appear()


        if 'origin' in kwargs:   # set the scale before model for correct corners
            setattr(self, 'origin', kwargs['origin'])
        if 'use_tags' in kwargs:
            setattr(self, 'use_tags', kwargs['use_tags'])

        if text != '':
            self.text = text

        for key, value in kwargs.items():
            if key == 'origin':
                continue
            setattr(self, key, value)




    def text_getter(self):
        t = ''
        y = 0
        if self.text_nodes:
            y = self.text_nodes[0].getY()

        for tn in self.text_nodes:
            if y != tn.getY():
                t += '\n'
                y = tn.getY()

            t += tn.node().text

        return t


    def text_setter(self, text): # set this to update the text.
        self.raw_text = text

        # clear stuff
        for img in self.images:
            destroy(img)
        self.images = []

        for tn in self.text_nodes:
            tn.remove_node()
        self.text_nodes = []
        if not text:
            return

        # check if using tags
        if (not self.use_tags
            # or self.text == self.start_tag or self.text == self.end_tag
            # or not self.start_tag in text or not self.end_tag in text
            ):

            self.create_text_section(text)
            self.align()
            return

        # parse tags
        text = self.start_tag + self.end_tag + str(text) # start with empty tag for alignment to work?
        sections = []
        section = ''
        tag = self.start_tag+'default'+self.end_tag
        temp_text_node = TextNode('temp_text_node')
        if self.font is not None:
            temp_text_node.setFont(self.font)
        x = 0
        y = 0

        i = 0
        while i < len(text):
            char = text[i]
            if char == '\n':
                sections.append([section, tag, x, y])
                section = ''
                y -= 1
                x = 0
                i += 1

            elif char == self.start_tag: # find tag
                sections.append([section, tag, x, y])
                x += temp_text_node.calcWidth(section)
                section = ''

                tag = ''
                done = False
                for j in range(len(text)-i):
                    tag += text[i+j]
                    if text[i+j] == self.end_tag and len(tag) > 0:
                        i += j+1
                        done = True
                        break
                if not done:
                    i += 1
            else:
                section += char
                i += 1

        sections.append([section, tag, x, y])

        for i, s in enumerate(sections):
            tag = s[1]
            # move the text after image one space right
            if tag.startswith(self.start_tag+'image:'):
                for f in sections:
                    if f[3] == s[3] and f[2] > s[2]:
                        f[2] += .5

                s[2] += .5

            self.create_text_section(text=s[0], tag=s[1], x=s[2], y=s[3])

        self.align()


    def create_text_section(self, text, tag='', x=0, y=0):
        # print(text, tag)
        self.text_node = TextNode('t')
        self.text_node_path = self.attachNewNode(self.text_node)
        try:
            self.text_node.setFont(self._font)
        except:
            pass    # default font


        if tag != '<>':
            tag = tag[1:-1]

            if tag.startswith('hsv('):   # set color based on numbers
                tag = tag[4:-1]
                hsb_values = tuple(float(e.strip()) for e in tag.split(','))
                self.current_color = color.hsv(*hsb_values)

            elif tag.startswith('rgb('):   # set color based on numbers
                tag = tag[4:-1]
                rgb_values = (float(e.strip()) for e in tag.split(','))
                self.current_color = color.rgba(*rgb_values)

            if tag.startswith('scale:'):
                scale = tag.split(':')[1]
                self.scale_override = float(scale)

            elif tag.startswith('image:'):
                texture_name = tag.split(':')[1]
                image = Entity(
                    parent=self.text_node_path,
                    shader=None,
                    name='inline_image',
                    model='quad',
                    texture=texture_name,
                    color=self.current_color,
                    # scale=self.scale_override,
                    # position=(x*self.size*self.scale_override, y*self.size*self.line_height),
                    origin=(.0, -.25),
                    add_to_scene_entities=False,
                    )
                if not image.texture:
                    destroy(image)
                else:
                    self.images.append(image)

            else:
                if tag in self.text_colors:
                    self.current_color = self.text_colors[tag]

        self.text_node_path.setScale(self.scale_override * self.size)
        self.text_node.setText(text)
        self.text_node.setTextColor(self.current_color)
        self.text_node.setPreserveTrailingWhitespace(True)

        self.text_node_path.setShader(self.shader._shader)
        for key, shader_input in self.shader.default_input.items():
            self.text_node_path.setShaderInput(key, shader_input)

        self.text_node_path.setPos(
            x * self.size * self.scale_override,
            (y * self.size * self.line_height) - .75 * self.size,
            0)
        self.text_nodes.append(self.text_node_path)

        return self.text_node

    def font_getter(self):
        return getattr(self, '_font', None)

    def font_setter(self, value):
        font_file_types = ('.ttf', '.otf') if '.' not in value else ('', )
        folders = (application.fonts_folder, application.asset_folder, application.internal_fonts_folder)
        font_file_path = _search_for_file(value, folders, font_file_types)

        if not font_file_path:
            print_warning('missing font:', value)

        if sys.platform == 'linux':
            font = FontPool.load_font(str(font_file_path))
        else:
            font = FontPool.load_font(str(font_file_path.name))

        if font:
            self._font = font
            self._font.clear()  # remove assertion warning
            self._font.setPixelsPerUnit(self.resolution)
            self._font.setLineHeight(self.line_height)
            if self.text:
                self.text = self.raw_text   # update tex


    def color_setter(self, value):
        self._color = value
        self.current_color = value
        self.text_colors['default'] = value
        for tn in self.text_nodes:
            tn.node().setTextColor(value)
        for img in self.images:
            img.color = value


    def shader_setter(self, value):
        self._shader = value
        if not hasattr(self, 'text_nodes'): # trying to set shader before init has finish, for example by using Entity.default_shader
            return

        for tn in self.text_nodes:
            tn.setShader(value._shader)
            for key, shader_input in value.default_input.items():
                tn.setShaderInput(key, shader_input)


    def line_height_getter(self):
        return getattr(self, '_line_height', 1)

    def line_height_setter(self, value):
        self._line_height = value
        if self.font is None:
            return
        if self.use_tags and self.text:
            self.text = self.raw_text
        else:
            self._font.setLineHeight(value)

    @property
    def width(self): # gets the width of the widest line.
        if not hasattr(self, 'text'):
            return 0

        temp_text_node = TextNode('temp')
        if self.font is None:
            print_warning('font not loaded, so Text.width may be wrong')
            return 1
        temp_text_node.setFont(self._font)

        longest_line_length = 0
        for line in self.text.split('\n'):
            longest_line_length = max(longest_line_length, temp_text_node.calcWidth(line))

        return longest_line_length * self.size


    @property
    def height(self): # gets the height of the text
        return (len(self.lines) * self.line_height * self.size)

    @property
    def lines(self):
        return self.text.splitlines()

    def resolution_getter(self):
        return self._font.getPixelsPerUnit()

    def resolution_setter(self, value):
        if self.font is None:
            return
        self._font.setPixelsPerUnit(value)


    def wordwrap_setter(self, value):   # set this to make the text wrap after a certain number of characters.
        self._wordwrap = value
        if not value:
            return

        new_text = ''
        for line in self.raw_text.split('\n'):
            x = 0
            for word in line.split(' '):
                clean_string = re.sub('<.*?>', '', word)
                x += len(clean_string) + 1
                # print('w:', word, 'len:', len(clean_string), 'clean str:', clean_string)

                if x >= value:
                    new_text += '\n'
                    x = 0

                new_text += word + ' '

            new_text += '\n'

        self.text = new_text


    def origin_setter(self, value):
        self._origin = value
        if self.text:
            self.text = self.raw_text


    def background_setter(self, value):
        if value is True:
            self.create_background()
        elif self.background_entity:
            from ursina.ursinastuff import destroy
            destroy(self.background_entity)


    def align(self):
        value = self.origin

        linewidths = [self.text_nodes[0].node().calcWidth(line) for line in self.lines]
        for tn in self.text_nodes:
            # center text horizontally
            # linenumber = abs(int(tn.getZ() / self.size / self.line_height))
            linenumber = abs(int(tn.getY() / self.size / self.line_height))
            linenumber = clamp(linenumber, 0, len(linewidths)-1)

            tn.setX(tn.getX() - (linewidths[linenumber] / 2 * self.size * tn.getScale()[0] / self.size))
            # tn.setX(tn.getX() - (linewidths[linenumber] / 2 * self.size))
            # add offset based on origin/value
            # x -= half line width * text node scale
            tn.setX(
                tn.getX() - (linewidths[linenumber] / 2 * value[0] * 2 * self.size) * tn.getScale()[0] / self.size
                )
            # center text vertically
            halfheight = len(linewidths) * self.line_height / 2
            # tn.setZ(tn.getZ() + (halfheight * self.size))
            tn.setY(tn.getY() + (halfheight * self.size))
            # add offset
            # tn.setZ(tn.getZ() - (halfheight * value[1] * 2 * self.size))
            tn.setY(tn.getY() - (halfheight * value[1] * 2 * self.size))


    def create_background(self, padding=size*2, radius=.1, color=ursina.color.black66):
        from ursina import Quad, destroy

        if self.background_entity:
            destroy(self.background_entity)

        self.background_entity = Entity(parent=self, z=.01)

        if isinstance(padding, (int, float, complex)):
            padding = (padding, padding)

        w, h = self.width + padding[0], self.height + padding[1]
        self.background_entity.x -= self.origin_x * self.width
        # if self.origin_x == .5:
        #     self._background.x += self.origin_x * self.width * 2
        self.background_entity.y -= self.origin_y * self.height

        self.background_entity.scale = (w,h)
        self.background_entity.model = Quad(radius=radius, aspect=w/h)
        # self.background_entity.model = 'quad'
        self.background_entity.color = color


    def appear(self, speed=.025):   # make the text animate in, one character at a time
        self.enabled = True
        # self.visible = True   # setting visible seems to reset the colors
        if self.appear_sequence:
            self.appear_sequence.finish()

        self.appear_sequence = Sequence()
        for tn in self.text_nodes:
            target_text = tn.node().getText()
            tn.node().setText('')
            new_text = ''

            for char in target_text:
                new_text += char
                self.appear_sequence.append(Wait(speed))
                self.appear_sequence.append(Func(tn.node().setText, new_text))

        self.appear_sequence.start()
        return self.appear_sequence


    def get_width(string, font=None):
        t = Text(string, add_to_scene_entities=False)
        if font:
            t.font = font
        w = t.width
        from ursina import destroy
        destroy(t)
        return w



if __name__ == '__main__':
    from ursina import *
    from ursina import Ursina, dedent, window
    app = Ursina()
    # Text.size = .001
    descr = dedent('''
        <red>Rainstorm<default> <red>Rainstorm<default>
        Summon a rain storm to deal 5 <blue>water<default> damage to everyone, test including yourself.
        1234 1234 1234 1234 1234 1234 2134 1234 1234 1234 1234 1234 2134 2134 1234 1234 1234 1234
        Lasts for 4 rounds.''').strip()

    # Text.default_font = 'VeraMono.ttf'
    # Text.default_font = 'consola.ttf'
    # color.text_color = color.lime
    Text.default_resolution = 1080 * Text.size
    test = Text(text=descr, wordwrap=30, scale=1)


    # test.align()
    # test = Text(descr)

    # test.text = ''
    # print(test.images)
  # print('\n', test.text, '\n\n')
    # test.font = 'VeraMono.ttf'
    # Text.font = 'VeraMono.ttf'
    # test.origin = (.5, .5)
    # test.origin = (0, 0)
    # test.wordwrap = 40

    # text = Text(text=descr, wordwrap=10, origin=(-.5,.5), y=.25, background=True)
    # Entity(parent=camera.ui, model='circle', scale=.05, color=color.yellow, y=text.y, z=-1)


    def input(key):
        if key == 'a':
            # test.appear(speed=.025)
            print('a')
            test.text = '<default><image:file_icon> <red><image:file_icon> test '
            print('by', test.text)

        if key == 'c':
            test.text = ''
    test.create_background()
    Sky(color=color.dark_gray)
    EditorCamera()
    window.fps_counter.enabled = False
    print('....', Text.get_width('yolo'))
    app.run()



My engine is focused on making the everyday game development process less frustrating and make the simple things easy to do. Not really that focused on the low level details, but on usability. I've basically stopped programming at work now because of how awful it is and rather program using my own framework at home.

Originally Unity stepped in the direction of making games less of a hassle, but they've just been straying further and further from that path in my option.
IMO DOTS is only really useful for a few types of games, but unnecessary for most. I only need a handful of movable/interactable object in my games. Even games that seems complex shouldn't actually be complex underneath.

Also, the iteration time is really bad. In Unity it's over one minute in a big project while in my engine it's close to 0.114 seconds. I really need to be able to work without disruption in order to be productive.

In Unreal, you're supposed to use Blueprint and C++. I don't think I have to say more about that. And the editor is not good either. I get why you say it's for visualizations now, because it's mostly focused on graphics. I would like to have that rendering engine though o_o

As an example, to create a GameObject in Unity at x=1 you write something like this. If I also wanted to add a mesh, god knows how much more code that would be.
```c#
var gameObject = new GameObject("name");
gameObject.transform.localPosition = new Vector3(1.0f, gameObject.transform.localPosition.y, gameObject.transform.localPosition.z);
```

In my engine I write:
```python
entity = Entity(x=1)
```



from ursina import *

app = Ursina()

# for y in range(tex.height):
#     skip = 0
#     for x in range(tex.width):
#
#         if skip > 0:
#             skip -= 1
#             continue
#
#         pixel = color.rgb(*tex.get_pixel(x,y))
#         # print(pixel)
#         if pixel == clear_color:
#             continue
#
#         # print(pixel)
#         b = color.brightness(tex.get_pixel(x,y))
#         # b_right = color.brightness(tex.get_pixel(x+1,y))
#         # b_up = color.brightness(tex.get_pixel(x,y+1))
#         # b_up_right = color.brightness(tex.get_pixel(x+1,y+1))
#
#         j = 0
#
#         for j in range(tex.width-x):
#             if color.rgb(*tex.get_pixel(x+j, y)) != pixel:
#                 break
#
#         verts.extend((
#             (x, b, y),
#             (x +1+j, b, y),
#             (x +1+j, b, y +1),
#             (x, b , y +1)
#             ))
#
#         tris.append((i, i+1, i+2, i+3))
#         # uvs.append((x/tex.width, y/tex.height))
#         # uvs.append(((x+skip) /tex.width, y/tex.height))
#         # uvs.append(((x+skip) /tex.width, (y+skip) /tex.height))
#         # uvs.append((x /tex.width, (y+skip) /tex.height))
#         # colors.extend((pixel, )*4)
#
#         i += 4
#         skip = j




def preview(tex, clear_color=color.white):
    # global preview_model
    print('.........')
    verts = list()
    colors = list()

    for y in range(tex.height):
        for x in range(tex.width):
            pixel = color.rgb(*tex.get_pixel(x,y))

            if pixel == clear_color:
                continue

            neighbors = (
                color.rgb(*tex.get_pixel(x,y+1)),
                color.rgb(*tex.get_pixel(x+1,y)),
                color.rgb(*tex.get_pixel(x,y-1)),
                color.rgb(*tex.get_pixel(x-1,y)),
                )

            if (pixel == neighbors[0]
                and pixel == neighbors[1]
                and pixel == neighbors[2]
                and pixel == neighbors[3]
                ):

                verts.append((x,y,0))
                colors.append(pixel)
                continue

            for z in range(16):
                verts.append((x,y,z))
                if z == 0:
                    colors.append(pixel)
                else:
                    colors.append(pixel.tint(-.2))

    preview_model.model = Mesh(vertices=verts, colors=colors, mode='point', thickness=32)
    preview_model.position =  (-tex.width/2*preview_model.scale_x, ) * 2     # center


def input(key):
    if held_keys['control'] and key == 'r':
        t = time.time()
        preview(tex)
        print('duration:', time.time() - t)


tex = load_texture('pot')
preview_model = Entity(scale=1 / max(tex.size) * 8)

preview(tex)


EditorCamera()
app.run()



import sys
sys.path.append("..")
from ursina import *

class NewScriptButton():

    def __init__(self):
        self.is_editor = True
        self.target = None


    def input(self, key):
        if key == 'left mouse down' and self.entity.hovered:
            self.target.enabled = not self.target.enabled



from ursina import *


class AudioExplorer(Panel):
    def __init__(self, **kwargs):
        super().__init__(parent=camera.ui, scale=.7, model=Quad(radius=.025),
            position=window.right+Vec2(-.7/2,0), color=color._16)

        self.sound_effects = {}
        self.current_sound_name = ''
        self.current_sound = None

        self.play_button = Button(parent=self, scale=.1, model='circle', color=color.azure, position=(-.5+.025,.5-.025,-.1), origin=(-.5,.5), text='>', on_click=self.play_sound)
        self.text_entity = Text(world_parent=self, text='None', position=self.play_button.position+Vec3(.12,-.05,-.1), origin=(-.5,0))
        self.copy_button = Button(parent=self, scale=(.1,.05), color=color.gray, position=(.5-.025,.5-.05,-.1), origin=(.5,.5), text='copy', on_click=Func(print, 'copy'))

        # original_default_font = Text.default_font
        # Text.default_font = 'VeraMono.ttf'

        for y, name in enumerate(('blip', 'boom', 'coin', 'hurt', 'jump', 'lose', 'powerup', 'teleport')):
            temp_x = 0
            for i in range(1,4):
                full_name = f'{name}_{i}'
                a = Audio(full_name, autoplay=False)
                #
                if not a.clip:
                    destroy(a)
                    continue

                destroy(a)
                #     def change_sound(full_name=full_name):
                #         self.current_sound_name = full_name
                #         self.play_sound()
                #
                #     self.sound_effects[full_name] = change_sound
                #
                # destroy(a)
                width = (Text.get_width(full_name) * 1.25) + .05
                # width = (.02*len(full_name))*1.25
                b = Button(parent=self, y=.5-.155-(y*.06*1.25), x=self.text_entity.x+temp_x, text=full_name,
                    scale=(width, .05*1.25),
                    color=color._64, z=-.1,
                    # text_color=color.light_gray
                    )
                b.model = Quad(radius=.5, aspect=b.scale_x/b.scale_y)
                b.origin = (-.5,.5)
                b.text_entity.scale *= .9
                temp_x += width + .01

                def change_sound(full_name=full_name):
                    self.current_sound_name = full_name
                    self.play_sound()
                b.on_click = change_sound


        self.pitch_slider = Slider(-12, 12, default=0, label='pitch', step=1, dynamic=True, on_value_changed=self.play_sound,
            parent=self, x=-.5+.04, y=-.5+.075, vertical=True, scale=1.5, z=-1)
        self.pitch_slider.bg.color = color._64
        self.volume_slider = Slider(0, 1, default=.5, label='volume', step=.05, dynamic=True, on_value_changed=self.play_sound,
            parent=self, x=-.5+.1, y=-.5+.075, vertical=True, scale=1.5, z=-1)
        self.volume_slider.bg.color = color._64

        # Text.default_font = original_default_font
        # bl = ButtonList(self.sound_effects, parent=self)
        # bl.model = 'wireframe_cube'
        # bl.origin = (-.5,.5)
        # self.bg = Entity(parent=self, model=Quad(.025), z=.1, scale=(.6,.65), origin_x=.5, x=.27, y=.215, collider='box', color=color.black66)


        for key, value in kwargs.items():
            setattr(self, key, value)


    def play_sound(self):
        if not self.current_sound_name:
            return

        new_pitch = pow(1 / 1.05946309436, -self.pitch_slider.value)
        new_pitch = round(new_pitch, 4)

        if self.current_sound:
            self.current_sound.stop()

        self.current_sound = Audio(self.current_sound_name, pitch=new_pitch, volume=self.volume_slider.value)

        # print(self.current_sound)
        self.text_entity.text = f"Audio('{self.current_sound_name}', pitch={new_pitch}, volume={round(self.volume_slider.value,2)})"


if __name__ == '__main__':
    app = Ursina()
    window.color = color.gold
    from time import perf_counter
    t = perf_counter()
    AudioExplorer()
    print('---', perf_counter() - t)

    bg = Sprite('shore', color=color.gray)
    app.run()



from ursina import Entity, Mesh, load_model, color, camera, Vec3, Vec2



class TilemapText(Entity):
    start_tag = '<'
    end_tag = '>'
    quad_model = None

    def __init__(self, text='', **kwargs):
        super().__init__(parent=camera.ui)

        if not TilemapText.quad_model:
            TilemapText.quad_model = load_model('quad')

        self.model = Mesh(vertices=[], uvs=[], colors=[])
        self.texture = 'Hack_square_64x64'
        self.scale = .025
        self.origin_x = -.5
        self.origin_y = .5
        self.line_height = 1
        self.character_spacing = .5

        self.use_tags = True
        self.start_tag = TilemapText.start_tag
        self.end_tag = TilemapText.end_tag
        self.text_colors = color.colors
        self.text_colors['default'] = color.text_color

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.current_color = self.text_colors['default']

        if text:
            self.text = text

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        # print('set text:', value)
        self._text = value
        self.model.vertices.clear()
        self.model.uvs.clear()
        self.model.colors.clear()

        parsing_tag = False
        lines = value.split('\n')

        for y, line in enumerate(lines):
            x = 0
            for char in line:
                if char == 'space':
                    x += 1
                    continue

                if self.use_tags:
                    if char == '<':
                        parsing_tag = True
                        tag = ''
                        continue

                    elif char == '>':
                        print('tag:', tag)
                        parsing_tag = False
                        if tag in self.text_colors:
                            self.current_color = self.text_colors[tag]
                        continue

                    elif parsing_tag:
                        tag += char
                        continue

                pos = ord(char)
                ascii_table_y = pos // 16
                ascii_table_x = pos - (ascii_table_y * 16)
                # print('x:', ascii_table_x, 'row:', ascii_table_y, 'char:', char)
                self.model.vertices += [Vec3(*v)
                    + Vec3(x*self.character_spacing, -y*self.line_height, 0)
                    + Vec3(.5,-.5,0) # offset quad
                    + Vec3(-len(line) * self.character_spacing * (.5+self.origin_x), len(lines) * self.line_height * (.5-self.origin_y), 0) # center
                    for v in TilemapText.quad_model.vertices]
                self.model.uvs += [(Vec2(*e)+Vec2(ascii_table_x,15-ascii_table_y))/16 for e in TilemapText.quad_model.uvs]
                self.model.colors += [self.current_color for i in range(len(TilemapText.quad_model.vertices))]
                x += 1

        self.model.generate()


    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = value
        if self.text:
            self.text = self._text


if __name__ == '__main__':
    from ursina import Ursina, dedent, EditorCamera
    app = Ursina()
    # descr = dedent('''
    #     <scale:1.5><orange>Rainstorm<default><scale:1>
    #     Summon a <azure>rain storm <default>to deal 5 <azure>water
    #
    #     damage <default>to <hsb(0,1,.7)>everyone, <default><image:file_icon> <red><image:file_icon> test <default>including <orange>yourself. <default>
    #     Lasts for 4 rounds.''').strip()
    descr = dedent('''
    123456765432

    2134
    4356
    6578

    ''').strip()
    # make_text(descr)
    # s = Sprite('acorn_font_tileset')
    # s.scale *= 3
    # s.texture_scale = (1/16, 1/16)
    # s.texture_offset = (1/16, (16-4)/16)
    Entity(model='quad', scale=.1, color=color.red)
    t = TilemapText(descr, scale=.05)
    # t.text = 'oisefjaeoifj\nopijefoieaj'


    EditorCamera()

    app.run()



'''
SDF text shader based on: https://github.com/wezu/sdf_text
License: Copyright (c) 2018, wezu (wezu.dev@gmail.com)

Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
'''

from ursina import color
from ursina.shader import Shader
from ursina.vec2 import Vec2

text_shader = Shader(name='text_shader', language=Shader.GLSL,
vertex='''#version 130


uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 uvs;

in vec4 p3d_Color;
out vec4 vertex_color;


void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    uvs = p3d_MultiTexCoord0;
    vertex_color = p3d_Color;
}
''',

fragment='''
#version 140

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
uniform vec4 outline_color;
uniform vec2 outline_offset;
uniform float outline_power;
in vec2 uvs;
out vec4 FINAL_COLOR;

in vec4 vertex_color;

void main() {
    float dist = texture(p3d_Texture0, uvs).a;
    vec2 width = vec2(0.5-fwidth(dist), 0.5+fwidth(dist));
    float alpha = smoothstep(width.x, width.y, dist);
    //supersampling
    float scale = 0.354; // half of 1/sqrt2 - value from internet(???)
    vec2 duv = scale * (dFdx(uvs) + dFdy(uvs));
    vec4 box = vec4(uvs-duv, uvs+duv);
    alpha +=0.5*(smoothstep(width.x, width.y, texture(p3d_Texture0, box.xy).a)
            +smoothstep(width.x, width.y, texture(p3d_Texture0, box.zw).a)
            +smoothstep(width.x, width.y, texture(p3d_Texture0, box.xw).a)
            +smoothstep(width.x, width.y, texture(p3d_Texture0, box.zy).a));
    alpha /= 3.0; //weighted average, 1*1 + 4*0.5 = 3, so divide by 3
    //outline
    float outline = pow(texture(p3d_Texture0, uvs-outline_offset).a, outline_power);
    FINAL_COLOR = mix(vec4(vertex_color.rgb, outline_color.a * outline), vertex_color, alpha);
}
''',

default_input = {
    'outline_color': color.white,
    'outline_offset': Vec2(10,10),
    'outline_power': 1.0,
}
)

if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    # window.color = color._16

    text_entity = Text('some <orange>COOL <default> text here', parent=scene, scale=25)
    text_shader.compile()

    for tn in text_entity.text_nodes:
        tn.setShader(text_shader._shader)
        for key, value in text_shader.default_input.items():
            tn.setShaderInput(key, value)
            print(tn.getColorScale())

    EditorCamera()

    app.run()



import os
import glob
import platform
import subprocess
from copy import copy, deepcopy
from pathlib import Path
from ursina.mesh import Mesh
from ursina import application
from panda3d.core import CullFaceAttrib
from time import perf_counter
from ursina.string_utilities import print_info, print_warning
from ursina import color

imported_meshes = dict()
blender_scenes = dict()

def load_model(name, path=application.asset_folder, file_types=('.bam', '.ursinamesh', '.obj', '.glb', '.gltf', '.blend'), use_deepcopy=False):
    if not isinstance(name, str):
        raise TypeError(f"Argument save must be of type str, not {type(str)}")

    if '.' in name:
        full_name = name
        name = full_name.split('.')[0]
        file_types = ('.' + full_name.split('.',1)[1],)

    if name in imported_meshes:
        # print('load cached model', name)
        try:
            if not use_deepcopy:
                instance = copy(imported_meshes[name])
            else:
                instance = deepcopy(imported_meshes[name])

            instance.clearTexture()
            return instance

        except:
            pass

    for filetype in file_types:
        # warning: glob is case-insensitive on windows, so m.path will be all lowercase
        for filename in path.glob(f'**/{name}{filetype}'):
            if filetype == '.bam':
                print_info('loading bam')
                return loader.loadModel(filename)

            if filetype == '.ursinamesh':
                try:
                    with open(filename) as f:
                        m = eval(f.read())
                        m.path = filename
                        m.name = name
                        imported_meshes[name] = m
                        return m
                except:
                    print_warning('invalid ursinamesh file:', filename)


            if filetype == '.obj':
                # print('found obj', filename)
                # m = loader.loadModel(filename)
                # m.setAttrib(CullFaceAttrib.make(CullFaceAttrib.MCullCounterClockwise))
                m = obj_to_ursinamesh(path=path, name=name, return_mesh=True)
                m.path = filename
                m.name = name
                imported_meshes[name] = m
                return m

            elif filetype == '.blend':
                print_info('found blend file:', filename)
                if blend_to_obj(path=path, name=name):
                    # obj_to_ursinamesh(name=name)
                    return load_model(name, path)

            else:
                try:
                    return loader.loadModel(filename)
                except:
                    pass

    return None


# find blender installations
if application.development_mode:

    if platform.system() == 'Windows':
        # get blender path by getting default program for '.blend' file extension
        import shlex
        import winreg

        try:
            class_root = winreg.QueryValue(winreg.HKEY_CLASSES_ROOT, '.blend')
            with winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, fr'{class_root}\shell\open\command') as key:
                command = winreg.QueryValueEx(key, '')[0]
                default_blender = shlex.split(command)[0]
                default_blender = Path(default_blender)
                application.blender_paths['default'] = default_blender
                blender_foundation_directory = default_blender.parent.parent

                for blender_installation in blender_foundation_directory.glob('*'):
                    first_folder = tuple(blender_installation.glob('*'))[0] # version
                    version_name = first_folder.name[:3]
                    blender_path = list(blender_installation.glob('blender.exe'))[0]
                    if application.blender_path['default'] == blender_path:
                        continue
                    application.blender_paths[version_name] = blender_path
        except:
            pass

    elif platform.system() == 'Linux':
        # Use "which" command to find blender
        which_process = subprocess.run(('which', 'blender'), stdout=subprocess.PIPE)
        if which_process.returncode == 0:
            blender_exec = which_process.stdout.decode().strip()
            application.blender_paths['default'] = blender_exec


def load_blender_scene(name, path=application.asset_folder, load=True, reload=False, skip_hidden=True, models_only=False, max_decimals=4, triangulate=True):
    scenes_folder = Path(application.asset_folder / 'scenes')
    if not scenes_folder.exists():
        scenes_folder.mkdir()

    out_file_path = scenes_folder / f'{name}.py'
    # print('loading:', out_file_path)
    if reload or not out_file_path.exists():
        print_info('reload:')
        t = perf_counter()
        blend_file = tuple(path.glob(f'**/{name}.blend'))
        if not blend_file:
            raise ValueError('no blender file found at:', path / name)

        blend_file = blend_file[0]
        print_info('loading blender scene:', blend_file, '-->', out_file_path)
        blender = get_blender(blend_file)
        script_path = application.internal_scripts_folder / '_blender_scene_to_ursina.py'

        args = [
            get_blender(blend_file),
            blend_file,
            '--background',
            '--python',
            application.internal_scripts_folder / '_blender_scene_to_ursina.py',
            out_file_path,
            '--skip_hidden' if skip_hidden else '',
            '--models_only' if models_only else '',
            '--triangulate' if models_only else '',
            f'--decimals={max_decimals}',
        ]

        subprocess.run(args)


    with open(out_file_path) as f:
        file_content = f.read()
        loc = {}
        exec(file_content, globals(), loc)

        if models_only:
            blender_scenes[name] = loc['meshes']
            return loc['meshes']

        blender_scenes[name] = loc['scene_parent']
        return loc['scene_parent']



def get_blender(blend_file):    # try to get a matching blender version in case we have multiple blender version installed
    if not application.blender_paths:
        raise Exception('error: trying to load .blend file, but no blender installation was found. blender_paths:', application.blender_paths)

    if len(application.blender_paths) == 1:
        return application.blender_paths['default']

    with open(blend_file, 'rb') as f:
        try:
            blender_version_number = (f.read(12).decode("utf-8"))[-3:]   # get version from start of .blend file e.g. 'BLENDER-v280'
            blender_version_number = blender_version_number[0] + '.' + blender_version_number[1:2]
            print_info('blender_version:', blender_version_number)
            if blender_version_number in application.blender_paths:
                return application.blender_paths[blender_version_number]

            print_info('using default blender version')
            return application.blender_paths['default']
        except:
            print_info('using default blender version')
            return application.blender_paths['default']


def blend_to_obj(path=None, outpath=application.models_compressed_folder, name='*'):

    if not application.models_compressed_folder.exists():
        application.models_compressed_folder.mkdir()

    export_script_path = application.internal_scripts_folder / '_blend_export.py'
    exported = []

    for blend_file in path.glob(f'**/{name}.blend'):
        blender = get_blender(blend_file)

        out_file_path = outpath / (blend_file.stem + '.obj')
        print_info('converting .blend file to .obj:', blend_file, '-->', out_file_path, 'using:', blender)

        if platform.system() == 'Windows':
            subprocess.call(f'''"{blender}" "{blend_file}" --background --python "{export_script_path}" "{out_file_path}"''')
        else:
            subprocess.run((blender, blend_file, '--background', '--python', export_script_path, out_file_path))

        exported.append(blend_file)

    return exported


def obj_to_ursinamesh(path=application.models_compressed_folder, outpath=application.models_compressed_folder, name='*', return_mesh=True, save_to_file=False, delete_obj=False):

    if name.endswith('.obj'):
        name = name[:-4]

    for f in path.glob(f'**/{name}.obj'):
        print('read .obj at:', f)

        with f.open('r') as file:
            lines = file.readlines()

        verts = []
        tris = []

        uv_indices = []
        uvs = []
        norm_indices = []
        norms = []
        normals = [] # final normals made getting norms with norm_indices

        vertex_colors = []
        current_color = None
        current_texture = None
        mtl_data = None
        materials = {}
        meshes = []
        textures = []
        offset = 0

        # parse the obj file to a Mesh
        for l in lines:
            if l.startswith('v '):
                vert = [float(v) for v in l[2:].strip().split(' ')]
                vert[0] = -vert[0]
                verts.append(tuple(vert))

            elif l.startswith('vn '):
                n = l[3:].strip().split(' ')
                norms.append(tuple(float(e) for e in n))

            elif l.startswith('vt '):
                uv = l[3:].strip()
                uv = uv.split(' ')
                uvs.append(tuple(float(e) for e in uv))

            elif l.startswith('f '):
                l = l[2:]
                l = l.split(' ')

                try:
                    tri = tuple(int(t.split('/')[0])-1 for t in l if t != '\n')
                except:
                    print_warning('error in obj file line:', i, ':', l)
                    return
                if len(tri) == 3:
                    tris.extend(tri)
                    if current_color:
                        vertex_colors.extend([current_color for i in range(3)])

                elif len(tri) == 4:
                    tris.extend((tri[0], tri[1], tri[2], tri[2], tri[3], tri[0]))
                    if current_color:
                        vertex_colors.extend([current_color for i in range(6)])

                else: # ngon
                    for i in range(1, len(tri)-1):
                        tris.extend((tri[i], tri[i+1], tri[0]))
                    if current_color:
                        vertex_colors.extend([current_color for i in range(len(tri))])

                try:
                    uv = tuple(int(t.split('/')[1])-1 for t in l)
                    if len(uv) == 3:
                        uv_indices.extend(uv)
                    elif len(uv) == 4:
                        uv_indices.extend((uv[0], uv[1], uv[2], uv[2], uv[3], uv[0]))
                    else: # ngon
                        for i in range(1, len(uv)-1):
                            uv_indices.extend((uv[i], uv[i+1], uv[0]))
                except: # if no uvs
                    pass

                try:
                    n = tuple(int(t.split('/')[2])-1 for t in l)
                    if len(n) == 3:
                        norm_indices.extend(n)
                    elif len(uv) == 4:
                        norm_indices.extend((n[0], n[1], n[2], n[2], n[3], n[0]))
                    else: # ngon
                        for i in range(1, len(n)-1):
                            norm_indices.extend((n[i], n[i+1], n[0]))
                except: # if no normals
                    pass

            elif l.startswith('mtllib '):    # load mtl file
                mtl_file_name = Path(str(f).rstrip('.obj') + '.mtl')
                # print('found mtl:', mtl_file_name)
                if mtl_file_name.exists():
                    with open(mtl_file_name) as mtl_file:
                        mtl_data = mtl_file.readlines()

                        for i in range(len(mtl_data)-1):
                            if mtl_data[i].startswith('newmtl '):
                                material_name = mtl_data[i].strip()[7:] # remove 'newmtl '
                                materials[material_name] = {'color':None, 'texture':None}

                                for j in range(i+1, min(i+10, len(mtl_data))):
                                    if mtl_data[j].startswith('newmtl'):
                                        break
                                    if mtl_data[j].startswith('Kd '):
                                        # print('found color')
                                        material_color = [float(e) for e in mtl_data[j].strip()[3:].split(' ')]
                                        materials[material_name]['color'] = *material_color, 1
                                        continue
                                    if mtl_data[j].startswith('map_Kd '):
                                        materials[material_name]['texture'] = mtl_data[j][7:].strip()
                                        # print('found texture:', mtl_data[j][7:])
                                        continue


            elif l.startswith('usemtl ') and mtl_data: # apply material color
                material_name = l[7:].strip()    # remove 'usemtl '
                if material_name in materials:
                    current_color = materials[material_name]['color']
                    # print('------',current_texture, materials[material_name]['texture'])

            elif l.startswith('o ') and mtl_data:
                # if current_texture != materials[material_name]['texture'] and tris:  # if we change texture, make a new Mesh
                current_texture = materials[material_name]['texture']
                textures.append(current_texture)
                vertices = [verts[t] for t in tris]

                # print('NEW MAEH')
                m = Mesh(vertices=vertices,normals=normals, uvs=[uvs[uid] for uid in uv_indices], colors=vertex_colors)
                meshes.append(m)
                offset += len(tris)
                # verts.clear()
                tris.clear()
                normals.clear()
                uvs.clear()
                uv_indices.clear()
                vertex_colors.clear()


        if norms: # make sure we have normals and not just normal indices (weird edge case).
            normals = [(-norms[nid][0], norms[nid][1], norms[nid][2]) for nid in norm_indices]

        if return_mesh:
            if not meshes:  # return one mesh
                return Mesh(vertices=[verts[t] for t in tris], normals=normals, uvs=[uvs[uid] for uid in uv_indices], colors=vertex_colors)

            # return multiple meshes with different textures
            from ursina import Entity
            mesh_parent = Entity()
            for i, m in enumerate(meshes):
                print(i, textures[i])
                Entity(parent=mesh_parent, model=m, texture=textures[i])

            return mesh_parent

        meshstring = ''
        meshstring += 'Mesh('

        meshstring += '\nvertices='
        meshstring += str(tuple(verts[t] for t in tris))

        if vertex_colors:
            meshstring += '\ncolors='
            meshstring += str(tuple(col for col in vertex_colors))

        if uv_indices:
            meshstring += ', \nuvs='
            meshstring += str(tuple(uvs[uid] for uid in uv_indices))

        if normals:
            meshstring += ', \nnormals='
            meshstring += str(normals)

        meshstring += ''', \nmode='triangle')'''

        if not save_to_file:
            return meshstring

        outfilepath = outpath / (os.path.splitext(f)[0] + '.ursinamesh')
        with open(outfilepath, 'w') as file:
            file.write(meshstring)

        if delete_obj:
            os.remove(filepath)

        print_info('saved ursinamesh to:', outfilepath)

# faster, but does not apply modifiers
def blend_to_obj_fast(model_name=None, write_to_disk=False):
    print_info('find models')
    from tinyblend import BlenderFile
    application.models_compressed_folder.mkdir(parents=True, exist_ok=True)

    files = os.listdir(application.models_folder)
    compressed_files = os.listdir(application.models_compressed_folder)

    for f in files:
        if f.endswith('.blend'):
            # print('f:', application.models_compressed_folder + '/' + f)
            print('compress______', f)
            blend = BlenderFile(application.models_folder + '/' + f)
            number_of_objects = len(blend.list('Object'))

            for o in blend.list('Object'):
                if not o.data.mvert:
                    continue
                # print(o.id.name.decode("utf-8", "strict"))
                object_name = o.id.name.decode( "utf-8").replace(".", "_")[2:]
                object_name = object_name.split('\0', 1)[0]
                print('name:', object_name)

                verts = [v.co for v in o.data.mvert]
                verts = tuple(verts)

                file_content = 'Mesh(' + str(verts)

                file_name = ''.join([f.split('.')[0], '.ursinamesh'])
                if number_of_objects > 1:
                    file_name = ''.join([f.split('.')[0], '_', object_name, '.ursinamesh'])
                file_path = os.path.join(application.models_compressed_folder, file_name)
                print(file_path)

                tris = tuple(triindex.v for triindex in o.data.mloop)
                flippedtris = []
                for i in range(0, len(tris)-3, 3):
                    flippedtris.append(tris[i+2])
                    flippedtris.append(tris[i+1])
                    flippedtris.append(tris[i+0])

                file_content += ', triangles=' + str(flippedtris)

                if o.data.mloopuv:
                    uvs = tuple(v.uv for v in o.data.mloopuv)
                    file_content += ', uvs=' + str(uvs)

                file_content += ''', mode='triangle')'''

                if write_to_disk:
                    with open(file_path, 'w') as file:
                        file.write(file_content)

                return file_content

def ursina_mesh_to_obj(mesh, name='', out_path=application.models_compressed_folder, max_decimals=3):
    from ursina.string_utilities import camel_to_snake

    if not name:
        name = camel_to_snake(mesh.__class__.__name__)
    obj = 'o ' + name + '\n'


    for v in mesh.vertices:
        v = [round(e, max_decimals) for e in v]
        obj += f'v {v[0]} {v[1]} {v[2]}\n'

    if mesh.uvs:
        for uv in mesh.uvs:
            uv = [round(e, max_decimals) for e in uv]
            obj += f'vt {uv[0]} {uv[1]}\n'

    obj += 's off\n'

    if mesh.triangles:
        tris = mesh.triangles

        if isinstance(tris[0], tuple): # convert from tuples to flat
            new_tris = []
            for t in tris:
                if len(t) == 3:
                    new_tris.extend([t[0], t[1], t[2]])
                elif len(t) == 4: # turn quad into tris
                    new_tris.extend([t[0], t[1], t[2], t[2], t[3], t[0]])

            tris = new_tris


    if mesh.mode == 'ngon':
        tris = []
        for i in range(1, len(mesh.vertices)-1):
            tris.extend((i, i+1, 0))


    # tris must be a list of indices
    for i, t in enumerate(tris):
        if i % 3 == 0:
            obj += '\nf '
        obj += str(t+1)
        if mesh.uvs:
            obj += '/'+str(t+1)
        obj += ' '


    # print(obj)
    with open(out_path / (name + '.obj'), 'w') as f:
        f.write(obj)
        print_info('saved obj:', out_path / (name + '.obj'))



def compress_internal():
    blend_to_obj(application.internal_models_folder)
    obj_to_ursinamesh(
        application.internal_models_compressed_folder,
        application.internal_models_compressed_folder,
        return_mesh=False, save_to_file=True, delete_obj=True
        )


if __name__ == '__main__':
    # compress_internal()
    from ursina import *
    app = Ursina()
    # print('imported_meshes:\n', imported_meshes)
    # Entity(model='quad').model.save('quad.bam')
    # m = obj_to_ursinamesh(path=application.asset_folder.parent / 'samples', name='procedural_rock_0')
    # Entity(model=m)
    # EditorCamera()


    # application.asset_folder = application.asset_folder.parent / 'samples'
    # application.asset_folder = Path(r'C:\\Users\\Petter\\Downloads\\')
    # Entity(model='c1a0')
    # from ursina.shaders import lit_with_shadows_shader
    # Entity.default_shader = lit_with_shadows_shader
    # Entity(model='race')
    # Entity(model='ambulance', x=1.5)

    application.asset_folder = Path(r'''C:\Users\Petter\Downloads''')
    t = perf_counter()
    Entity(model='untitled')
    print('-------', perf_counter() - t)
    # ground = Entity(model='plane', scale=10, texture='brick', texture_scale=Vec2(4))
    # DirectionalLight()

    # blender_scene = load_blender_scene(path=application.asset_folder, name='desert', reload=True)
    # blender_scene = load_blender_scene(path=application.asset_folder, name='blender_level_editor_test_scene_2')
    # print('-------', time.time() - t)

    # print('--------', blender_scene.children)
    # for e in blender_scene.children:
    #     # e.color = color.random_color()
    #     e.shader = rim_shader
    #     e.texture='matcap_4'
    #
    #
    # blender_scene.Plane_002.collider = 'mesh'
    # from ursina.prefabs.first_person_controller import FirstPersonController
    # player = FirstPersonController()

    # def input(key):
    #     if key == '+':
    #         for e in blender_scene.children:
    #             e.texture_scale = Vec2(e.texture_scale[0], e.texture_scale[1]+.1)
    #     if key == '-':
    #         for e in blender_scene.children:
    #             e.texture_scale = Vec2(e.texture_scale[0], e.texture_scale[1]-.1)
    #         print(blender_scene.children[0].texture_scale)
    #
    EditorCamera()
    Sky(texture='sky_sunset')
    # def update():
    #     blender_scene.Cube.x += (held_keys['d'] - held_keys['a']) * time.dt * 10
    # application.asset_folder = Path('/home/poke/programming/ursina/samples/')
    # Entity(model='ambulance', x=-3)
    # # application.asset_folder = Path(Path('.').parent / 'samples' / 'models').resolve()
    # # print('----------', application.asset_folder)
    # application.asset_folder = Path('/home/poke/Downloads/Zazen')
    # e = Entity(model='Machi_StageNum_366.obj')

    application.asset_folder = Path('/home/poke/Downloads/')
    e = Entity(model='lamp', scale=.1, double_sided=True)
    # for c in e.model.children:
    #     c.texture='StoneWall.png'

    app.run()
    # e = Entity(model=Cylinder(16))
    # ursina_mesh_to_obj(e.model, name='quad_export_test')



from ursina import *
from ursina.models.procedural.quad import Quad
from ursina.prefabs.button import Button


class Panel(Entity):

    def __init__(self, **kwargs):
        super().__init__()
        self.parent = camera.ui
        self.model = Quad()
        self.color = Button.default_color

        for key, value in kwargs.items():
            setattr(self, key, value)

if __name__ == '__main__':
    app = Ursina()
    p = Panel()
    app.run()



from ursina import *
import random


app = Ursina()

color.text_color = color.dark_text

names = ['Amy', 'Ruby', 'Tara', 'Ann', 'Samantha', 'Gary', 'Lee', 'Frank', 'Joe', 'Thomas']

random.seed(0)
data = dict()
for name in names:
    data[name] = random.randint(0, 100)

sliders = list()

for i, (name, value) in enumerate(data.items()):
    column = Button(
        parent = scene,
        name = name,
        model = 'cube',
        x = i - (len(names)/2),
        scale = (.5, value/50, .5),
        color = color.hsv(30*i, 1, .7),
        origin_y = -.5,
        text = name,
        tooltip = Tooltip('00', color=color.light_text) # to ensure uniform with
    )
    column.tooltip.text = str(value)
    column.text_entity.scale *= .4
    column.text_entity.world_y = column.world_y - .2
    column.text_entity.z = -.5
    column.text_entity.world_parent = scene
    column.text_entity.color = column.color.tint(-.5)

    slider = ThinSlider(text=name, min=0, max=100, default=value, x=-.65, y=(-i*.04*.75) - .15, step=1, dynamic=True)
    slider.scale *= .75
    sliders.append(slider)

    def on_slider_changed(slider=slider, column=column):
        column.scale_y = slider.value/50
        column.tooltip.text = str(slider.value)

    slider.on_value_changed = on_slider_changed


randomize_button = Button(position=(-.66,-.45), origin=(-.5,.5), color=color.dark_gray, text='<white>Randomize!', scale=(.25, .05))
randomize_button.scale *= .75

def randomize():
    for s in sliders:
        s.value = random.randint(0,100)
        s.on_value_changed()

randomize_button.on_click = randomize



window.color=color.light_gray.tint(.1)
window.fps_counter.enabled = False
window.exit_button.visible = False
camera.orthographic = True
camera.fov = 8
EditorCamera()

app.run()



import os
from panda3d.bullet import BulletDebugNode
from types import MethodType
# import debugwindow

from ursina import *
from ursina.internal_prefabs.transform_gizmo import TransformGizmo
from ursina.internal_prefabs.hierarchy_panel import HierarchyPanel
from ursina.internal_prefabs.inspector import Inspector
from ursina.internal_scripts.editor_camera import EditorCamera




class Editor(Entity):

    def __init__(self):
        super().__init__()
        self.name = 'editor'
        self.is_editor = True
        self.parent = scene.ui

        raycaster.model = 'box'
        raycaster.color = color.yellow
        raycaster.position = (5,0,5)
        # raycaster.parent = self
        print(type(raycaster))
        self.trash = NodePath('trash')
        self.selection = list()

        self.editor_camera = self.add_script(EditorCamera())


        self.transform_gizmo = TransformGizmo()
        self.transform_gizmo.name = 'transform_gizmo'
        self.transform_gizmo.is_editor = True
        self.transform_gizmo.parent = render


        self.grid = Entity()
        self.grid.model = 'cube'
        self.grid.is_editor = True
        self.grid.name = 'grid_x'
        self.grid.parent = render
        self.grid.position = (0, 0, 0)
        self.grid.scale = (10, 0, .02)
        self.grid.color = color.orange

        self.grid = Entity()
        self.grid.model = 'cube'
        self.grid.is_editor = True
        self.grid.name = 'grid_z'
        self.grid.parent = render
        self.grid.position = (0, 0, 0)
        self.grid.rotation = (0, 90, 0)
        self.grid.scale = (10, 0, .02)
        self.grid.color = color.lime

# top menu
        self.top_menu = Entity()
        self.top_menu.parent = self
        self.top_menu.position = window.top
        self.layout_group = self.top_menu.add_script('grid_layout')
        self.layout_group.origin = (0, .5)
        self.layout_group.spacing = (.001, 0)

# play button
        self.play_button = EditorButton()
        self.play_button.parent = self.top_menu
        self.play_button.origin = (0, .5)
        self.play_button.name = 'play_button'
        self.play_button.scale = (.1, .05)
        self.play_button.text = 'play'
        # self.menu_toggler = self.play_button.add_script('menu_toggler')

        self.pause_button = EditorButton()
        self.pause_button.parent = self.top_menu
        self.pause_button.origin = (0, .5)
        self.pause_button.name = 'pause_button'
        self.pause_button.scale = (.1, .05)
        self.pause_button.text = 'pause'
        # self.menu_toggler = self.play_button.add_script('menu_toggler')

        self.layout_group.update_grid()

# load menu
        self.load_menu_parent = Entity()
        self.load_menu_parent.parent = self
        self.load_menu_parent.position = window.right
        self.load_menu_parent.y -= .1
        self.layout_group = self.load_menu_parent.add_script('grid_layout')
        self.layout_group.origin = (.5, .5)
        self.layout_group.max_x = 1
        self.layout_group.spacing = (0, .001)

        button_names = ('internal\nscenes', 'scenes', 'internal\nprefabs', 'prefabs', 'models', 'primitives', 'sprites')
        button_paths = (
            # application.internal_scenes_folder,
            application.scenes_folder,
            # application.internal_prefabs_folder,
            application.prefabs_folder,
            application.models_folder,
            application.internal_models_folder,
            application.textures_compressed_folder
            )
        file_types = (('.py'), ('.py'), ('.py'), ('.py'), ('.egg'), ('.egg'), ('.png', '.jpg', '.gif'))
        button_types = (
            'load_scene_button',
            'load_scene_button',
            'load_prefab_button',
            'load_prefab_button',
            'load_model_button',
            'load_model_button',
            'load_sprite_button'
            )

        for i in range(len(button_names)):
            b = EditorButton()
            b.parent = self.load_menu_parent
            b.name = button_names[i]
            b.origin = (.5, 0)
            b.scale = (.1, .05)
            b.text = button_names[i]
            menu_toggler = b.add_script('menu_toggler')
            b.add_script('open_in_file_explorer')
            b.open_in_file_explorer.path = button_paths[i]

            filebrowser = FileBrowser()
            filebrowser.parent = self
            filebrowser.close_button.enabled = False
            # filebrowser.enabled = False
            filebrowser.file_types = file_types[i]
            filebrowser.path = button_paths[i]
            filebrowser.button_type = button_types[i]
            # print('ffffffffffffffffffff', filebrowser)
            menu_toggler.target = filebrowser

        self.layout_group.update_grid()


# entity list
        self.hierarchy_panel = HierarchyPanel()
        self.hierarchy_panel.parent = self
        self.hierarchy_panel.populate()

        self.hierarchy_panel_header = EditorButton()
        self.hierarchy_panel_header.parent = self
        self.hierarchy_panel_header.z = -2
        self.hierarchy_panel_header.color = (color.lime + color.black) / 2
        self.hierarchy_panel_header.position = window.top_left
        self.hierarchy_panel_header.origin = (-.5, .5)
        self.hierarchy_panel_header.scale = (.25, .025)
        self.hierarchy_panel_header.text = scene.name
        self.hierarchy_panel_header.text_entity.align = 'left'
        self.hierarchy_panel_header.text_entity.x = -.45
        self.hierarchy_panel_header.add_script('menu_toggler')
        self.hierarchy_panel_header.menu_toggler.target = self.hierarchy_panel

        self.save_scene_button = EditorButton()
        self.save_scene_button.parent = self.hierarchy_panel_header
        self.save_scene_button.color = color.green
        self.save_scene_button.text = 's'
        self.save_scene_button.origin = (0.5, 0)
        self.save_scene_button.position = (1, -.5, -3)
        self.save_scene_button.scale_x = .15
        self.save_scene_button.add_script('save_scene_button')


# inspector
        self.inspector = Inspector()
        self.inspector.parent = self
#
# view front
        self.toggle_button = EditorButton()
        self.toggle_button.is_editor = True
        self.toggle_button.parent = self
        self.toggle_button.name = 'toggle_button'
        self.toggle_button.origin = (0, .5)
        self.toggle_button.position = window.top_right
        self.toggle_button.x -= .1
        self.toggle_button.scale = (.1, .05)
        self.toggle_button.text = 'front'
        self.toggle_button.add_script('toggle_sideview')



        # from panda3d.core import DirectionalLight
        # from panda3d.core import VBase4
        # light = DirectionalLight('light')
        # light.setColor(VBase4(1, 1, 1, 1))
        # dlnp = render.attachNewNode(light)
        # dlnp.setHpr(0, -60, 60)
        # # dlnp.setPos(0, -10, 10)
        # dlnp.setPos(0, 0, 32)
        # dlnp.setScale(100)
        # # dlnp.lookAt(0, 0, 0)
        # dlnp.node().getLens().setNearFar(.2, 100)
        # # dlnp.node().getLens().setFocalLength(100)
        #
        # render.setLight(dlnp)
        # # Use a 512x512 resolution shadow map
        # light.setShadowCaster(True, 2048, 2048)
        # # Enable the shader generator for the receiving nodes
        # # render.setShaderAuto()
        # light.showFrustum()

        # ground = Entity()
        # ground.model = 'quad'
        # ground.y = .1
        # ground.rotation_x = -90
        # ground.scale *= 10
        # ground.setShaderAuto()

        # cube = Entity()
        # cube.model = 'cube'
        # cube.origin = (0, -.5, 0)
        # cube.setShaderAuto()

        # self.t = Text()
        # self.t.text = 'test text'

        # self.text.color = color.smoke
        # self.text.parent = scene.ui


#         self.text_editor = Entity()
#         self.text_editor.parent = self
#         self.text_bg = EditorButton()
#         self.text_bg.parent = self.text_editor
#         self.text_bg.scale = (.6, .8)
#         self.text_bg.button_script._highlight_color = self.text_bg.button_script.color
#

        # self.compress_textures()
        # self.blend_to_obj()


    def update(self):
        for s in self.scripts:
            if hasattr(s, 'update'):
                s.update(dt)

    def input(self, key):
        if key == 'l':
            for e in scene.entities:
                if e.has_ancestor(scene):
                    print(e.name)

        if key == 'tab':
            self.enabled = not self.enabled
            if self.enabled:
                print('enable editor')
            else:
                print('disable editor')

        if not self.enabled:
            return

        for s in self.scripts:
            if hasattr(s, 'input'):
                s.input(key)
        # if key == 'l':
        #     render.setShaderAuto()
        #     print('set shader auto')

        if held_keys['control'] and key == 'z':
            undo.stack().undo()
        if held_keys['control'] and key == 'y':
            undo.stack().redo()


        if key == 'n':
            scene.new()
        if key == 'p':
            print('p')
            e = load_scene('cube_1')
            # e = load_script('cube_1')
            e.parent = render
            print('loaded')

        if key == 'h':
            self.show_colliders = not self.show_colliders
            if self.show_colliders:
                self.debugNP.show()
            else:
                self.debugNP.hide()


        # if self.enabled:
        #     self.editor_camera_script.input(key)



    def on_disable(self):
        self.transform_gizmo.enabled = False
        self.grid.enabled = False
        # disable editor
        self.editor_camera_script.position = camera.position
        camera.wrtReparentTo(render)
        for e in scene.entities:
            if e.is_editor:
                continue
            try:
                e.editor_collider.stash()
                e.collider.unstash()
            except:
                pass
            # try:
            #     e.start()
            # except:
            #     pass
            for s in e.scripts:
                try:
                    s.start()
                except:
                    pass


    def on_enable(self):
        # enable editor
        # if self.enabled:
        try:
            camera.wrtReparentTo(self.editor_camera.camera_pivot)
            camera.position = self.editor_camera.position
        except:
            pass

        for e in scene.entities:
            if e.is_editor or e is scene:
                continue

            try:
                e.editor_collider = 'box'
                e.collider.stash()
            except:
                pass
            try:
                e.stop()
            except:
                pass
            for s in e.scripts:
                try:
                    s.stop()
                except:
                    pass

        try:
            self.transform_gizmo.enabled = True
            self.grid.enabled = True
            mouse.locked = False
        except:
            print('editor not initialized')


    def blend_to_obj(self):
        subprocess.call(r'''"C:\Program Files\Blender Foundation\Blender\blender.exe" "D:\UnityProjects\pandagame\ursina\internal_models\cube.blend" --background --python "D:\UnityProjects\pandagame\ursina\internal_scripts\blend_export.py"''')
        return
        from tinyblend import BlenderFile
        from os.path import dirname
        files = os.listdir(application.models_folder)
        compressed_files = os.listdir(application.models_compressed_folder)
        # print(files)
        texture_dir = os.path.join(
            dirname(dirname(dirname(os.path.abspath(__file__)))),
            'textures'
            )

        for f in files:
            if f.endswith('.blend'):
                # print('f:', application.models_compressed_folder + '/' + f)
                print('______', f)
                blend = BlenderFile(application.models_folder + '/' + f)
                # objects = blend.list('Object')
                for o in blend.list('Object'):
                    # print(o.id.name.decode("utf-8", "strict"))
                    object_name = o.id.name.decode( "utf-8").replace(".", "_")[2:]
                    object_name = object_name.split('\0', 1)[0]
                    print('name:', object_name)
                    file_name = ''.join([f.split('.')[0], '_', object_name, '.egg'])
                    file_path = os.path.join(
                        str(Filename.toOsSpecific(application.models_compressed_folder)),
                        file_name
                    )
                    print(file_path)
                    print(len(o.data.mloop))

                    # for v in o.data.mvert:
                    #     print('vertex:', v.co)
                    # for e in o.data.medge:
                    #     print('edge:', e.v1, e.v2)
                    # for i in range(0, len(o.data.mloop), 3):
                    #     print('triangle:',
                    #         o.data.mloop[i].v,
                    #         o.data.mloop[i+1].v,
                    #         o.data.mloop[i+2].v
                    #     )
                    with open(file_path, 'w') as file:
                        file.write(
                            '<CoordinateSystem> { Z-up }\n'
                            + '<Group> ' + object_name + ' {\n'
                            +
'''<Transform> {
    <Matrix4> {
      1.0 0.0 0.0 0.0
      0.0 1.0 0.0 0.0
      0.0 0.0 1.0 0.0
      0.0 0.0 0.0 1.0
    }
  }'''

                            + '  <VertexPool> ' + object_name + ' {\n'
                        )
                        for i in range(len(o.data.mvert)):
                            file.write(
                                '    <Vertex> '
                                + str(i)
                                + ' {' + str(o.data.mvert[i].co[0])
                                + ' ' + str(o.data.mvert[i].co[1])
                                + ' ' + str(o.data.mvert[i].co[2]) + '\n'
                                + '      <UV> ORCO {' + '\n'
                                + '        ' + '0.000000 0.000000' + '\n'
                                + '      }\n'
                                + '    }\n'
                            )
                        file.write('  }\n')

                        for i in range(0, len(o.data.mloop)-2, 3):
                            file.write(''.join([
                                '  <Polygon> {\n',
                                '    <Normal> {', '0 0 0',  '}\n',
                                '    <VertexRef> { ',
                                str(o.data.mloop[i].v), ' ',
                                str(o.data.mloop[i+1].v), ' ',
                                str(o.data.mloop[i+2].v), ' <Ref> { ',
                                object_name, ' }}\n',
                                '  }\n'
                            ])
                            )

                        file.write('}')


if __name__ == '__main__':
    app = main.Ursina()
    scene.editor = Editor()
    app.run()



import sys
sys.path.append("..")
from ursina import *

class AskForSceneNameMenu(Entity):

    def __init__(self):
        super().__init__()
        self.parent = camera.ui
        self.model = 'quad'
        self.scale = (.25, .025)

        self.window_title = EditorButton()
        self.window_title.parent = self.model
        self.window_title.color = color.black
        self.window_title.text = 'save scene as...'

        self.window_title.text_entity.align = 'left'
        self.window_title.text_entity.x = -.45
        self.window_title.remove_script('editor_draggable')

        self.input_field = InputField()
        self.input_field.is_editor = True
        self.input_field.parent = self.model
        self.input_field.y = -1

        self.save_button = EditorButton()
        self.save_button.parent = self.model
        self.save_button.text = 'save'
        self.save_button.y = -2

        self.close_button = EditorButton()
        self.close_button.name = 'close_button'
        self.close_button.parent = self
        self.close_button.position = (0, 0, 10)
        self.close_button.scale *= 100
        self.close_button.color = color.black33
        self.close_button.highlight_color = color.black33
        menu_toggler = self.close_button.add_script('menu_toggler')
        menu_toggler.target = self


    def input(self, key):
        if self.save_button.hovered and key == 'left mouse down' or key == 'enter':
            print('save field:', self.input_field.text)
            if self.input_field.text != '':
                scene.name = self.input_field.text
                scene.editor.hierarchy_panel_header.text = self.input_field.text
                save_prefab(scene, application.scenes_folder)
                self.enabled = False
                print('saved scene:', 'scenes', scene_entity.name)

# class Save():
if __name__ == '__main__':
    app = Ursina()
    test = AskForSceneNameMenu()
    app.run()



from ursina import *

# print('--------', application.asset_folder)
application.asset_folder = Path(application.asset_folder / 'test')
app = Ursina()

Button('Here is some text').fit_to_text()
Text('Make a file called "ursina_settings.py" in the root folder and \nthe engine will run the code in it.', y=.4, z=-1, origin=(0,0))

app.run()
